import numpy as np
import pytest

from ssspy.linalg import invsqrtmh, sqrtmh

parameters_sources = [2]
parameters_channels = [3, 4]
parameters_frames = [32]
parameters_is_complex = [True, False]
parameters_is_flooring = [True, False]


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_channels", parameters_channels)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
def test_sqrtmh(n_sources: int, n_channels: int, n_frames: int, is_complex: bool):
    rng = np.random.default_rng(0)

    shape = (n_sources, n_channels, n_frames)

    if is_complex:
        x = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
        X = np.mean(x[:, :, np.newaxis, :] * x[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        x = rng.standard_normal(shape)
        X = np.mean(x[:, :, np.newaxis, :] * x[:, np.newaxis, :, :], axis=-1)

    X_sqrt = sqrtmh(X)

    assert np.allclose(X, X_sqrt @ X_sqrt)


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_channels", parameters_channels)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
@pytest.mark.parametrize("is_flooring", parameters_is_flooring)
def test_invsqrtmh(
    n_sources: int, n_channels: int, n_frames: int, is_complex: bool, is_flooring: bool
):
    rng = np.random.default_rng(0)

    shape = (n_sources, n_channels, n_frames)

    if is_complex:
        x = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
        X = np.mean(x[:, :, np.newaxis, :] * x[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        x = rng.standard_normal(shape)
        X = np.mean(x[:, :, np.newaxis, :] * x[:, np.newaxis, :, :], axis=-1)

    if is_flooring:
        X_invsqrt = invsqrtmh(X, flooring_fn=lambda x: np.maximum(x, 1e-10))
    else:
        X_invsqrt = invsqrtmh(X)

    X_sqrt = np.linalg.inv(X_invsqrt)

    assert np.allclose(X, X_sqrt @ X_sqrt)



import functools
from typing import Callable, List, Optional, Tuple, Union

import numpy as np

from ..algorithm import (
    MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS,
    PROJECTION_BACK_KEYWORDS,
    minimal_distortion_principle,
    projection_back,
)
from ..linalg.mean import gmeanmh
from ..linalg.quadratic import quadratic
from ..linalg.sqrtm import invsqrtmh, sqrtmh
from ..special.flooring import identity, max_flooring
from ..special.psd import to_psd
from ..utils.flooring import choose_flooring_fn
from ._update_spatial_model import update_by_block_decomposition_vcd
from .base import IterativeMethodBase

spatial_algorithms = ["FPI", "VCD"]
source_algorithms = ["EM", "MM"]
EPS = 1e-10


class IPSDTABase(IterativeMethodBase):
    r"""Base class of independent positive semidefinite tensor analysis (IPSDTA).

    Args:
        n_basis (int):
            Number of PSDTF bases.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["IPSDTABase"], None], List[Callable[["IPSDTABase"], None]]]
        ] = None,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        self.source_normalization: Optional[Union[bool, str]]

        super().__init__(callbacks=callbacks, record_loss=record_loss)

        self.n_basis = n_basis

        if flooring_fn is None:
            self.flooring_fn = identity
        else:
            self.flooring_fn = flooring_fn

        self.input = None
        self.scale_restoration = scale_restoration

        if reference_id is None and scale_restoration:
            raise ValueError("Specify 'reference_id' if scale_restoration=True.")
        else:
            self.reference_id = reference_id

        if rng is None:
            rng = np.random.default_rng()

        self.rng = rng

    def __call__(self, input: np.ndarray, n_iter: int = 100, **kwargs) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        super().__call__(n_iter=n_iter)

        if self.scale_restoration:
            self.restore_scale()

        self.output = self.separate(self.input, demix_filter=self.demix_filter)

        return self.output

    def __repr__(self) -> str:
        s = "IPSDTA("
        s += "n_basis={n_basis}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        **kwargs,
    ) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            kwargs:
                Keyword arguments to set as attributes of IPSDTA.
        """
        assert self.input is not None, "Specify data!"

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        n_channels, n_bins, n_frames = X.shape
        n_sources = n_channels  # n_channels == n_sources

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        if not hasattr(self, "demix_filter"):
            W = np.eye(n_sources, n_channels, dtype=np.complex128)
            W = np.tile(W, reps=(n_bins, 1, 1))
        else:
            if self.demix_filter is None:
                W = None
            else:
                # To avoid overwriting ``demix_filter`` given by keyword arguments.
                W = self.demix_filter.copy()

        self.demix_filter = W
        self.output = self.separate(X, demix_filter=W)

        self._init_psdtf(flooring_fn=flooring_fn, rng=self.rng)

    def _init_psdtf(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        r"""Initialize PSDTF.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        n_basis = self.n_basis
        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if rng is None:
            rng = np.random.default_rng()

        if not hasattr(self, "basis"):
            # should be positive semi-definite
            eye = np.eye(n_bins, dtype=np.complex128)
            rand = rng.random((n_sources, n_basis, n_bins))
            T = rand[..., np.newaxis] * eye
        else:
            # To avoid overwriting.
            T = self.basis.copy()

        if not hasattr(self, "activation"):
            V = rng.random((n_sources, n_basis, n_frames))
            V = flooring_fn(V)
        else:
            # To avoid overwriting.
            V = self.activation.copy()

        self.basis, self.activation = T, V

        if self.source_normalization:
            self.normalize_psdtf()

    def separate(self, input: np.ndarray, demix_filter: np.ndarray) -> np.ndarray:
        r"""Separate ``input`` using ``demixing_filter``.

        .. math::
            \boldsymbol{y}_{ij}
            = \boldsymbol{W}_{i}\boldsymbol{x}_{ij}

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            demix_filter (numpy.ndarray):
                The demixing filters to separate ``input``.
                The shape is (n_bins, n_sources, n_channels).

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        X, W = input, demix_filter
        Y = W @ X.transpose(1, 0, 2)
        output = Y.transpose(1, 0, 2)

        return output

    def reconstruct_psdtf(
        self,
        basis: np.ndarray,
        activation: np.ndarray,
        axis1: int = -2,
        axis2: int = -1,
    ) -> np.ndarray:
        r"""Reconstruct PSDTF.

        Args:
            basis (numpy.ndarray):
                Basis matrix.
                The shape is (n_sources, n_basis, n_bins, n_bins) if ``axis1=-1`` and ``axis2=-2``.
                Otherwise, (n_sources, n_bins, n_bins, n_basis).
            activation (numpy.ndarray):
                Activation matrix.
                The shape is (n_sources, n_basis, n_frames).
            axis1 (int):
                First axis of covariance matrix. Default: ``-2``.
            axis2 (int):
                Second axis of covariance matrix. Default: ``-1``.

        Returns:
            numpy.ndarray of reconstructed PSDTF.
            The shape is (n_sources, n_frames, n_bins, n_bins).
        """
        T, V = basis, activation
        n_dims = T.ndim

        axis1 = n_dims + axis1 if axis1 < 0 else axis1
        axis2 = n_dims + axis2 if axis2 < 0 else axis2

        assert (axis1 == 1 and axis2 == 2) or (axis1 == 2 and axis2 == 3)

        if axis1 == 1 and axis2 == 2:
            T = T.transpose(0, 3, 1, 2)

        R = np.sum(T[:, :, np.newaxis, :, :] * V[:, :, :, np.newaxis, np.newaxis], axis=1)
        R = to_psd(R, axis1=2, axis2=3)

        return R

    def update_once(self) -> None:
        r"""Update demixing filters once."""
        raise NotImplementedError("Implement 'update_once' method.")

    def normalize_psdtf(self) -> None:
        r"""Normalize PSDTF parameters."""
        source_normalization = self.source_normalization
        T, V = self.basis, self.activation

        assert source_normalization, "Set source_normalization."

        trace = np.trace(T, axis1=-2, axis2=-1).real
        T = T / trace[:, :, np.newaxis, np.newaxis]
        V = V * trace[:, :, np.newaxis]

        self.basis, self.activation = T, V

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        Returns:
            Computed loss.
        """
        raise NotImplementedError("Implement 'compute_loss' method.")

    def compute_logdet(self, demix_filter: np.ndarray) -> np.ndarray:
        r"""Compute log-determinant of demixing filter

        Args:
            demix_filter (numpy.ndarray):
                Demixing filters with shape of (n_bins, n_sources, n_channels).

        Returns:
            numpy.ndarray of computed log-determinant values.
        """
        _, logdet = np.linalg.slogdet(demix_filter)  # (n_bins,)

        return logdet

    def restore_scale(self) -> None:
        r"""Restore scale ambiguity.

        If ``self.scale_restoration="projection_back``, we use projection back technique.
        """
        scale_restoration = self.scale_restoration

        assert scale_restoration, "Set self.scale_restoration=True."

        if type(scale_restoration) is bool:
            scale_restoration = PROJECTION_BACK_KEYWORDS[0]

        if scale_restoration in PROJECTION_BACK_KEYWORDS:
            self.apply_projection_back()
        elif scale_restoration in MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS:
            self.apply_minimal_distortion_principle()
        else:
            raise ValueError("{} is not supported for scale restoration.".format(scale_restoration))

    def apply_projection_back(self) -> None:
        r"""Apply projection back technique to estimated spectrograms."""
        assert self.scale_restoration, "Set self.scale_restoration=True."

        X, W = self.input, self.demix_filter
        W_scaled = projection_back(W, reference_id=self.reference_id)
        Y_scaled = self.separate(X, demix_filter=W_scaled)

        self.output, self.demix_filter = Y_scaled, W_scaled

    def apply_minimal_distortion_principle(self) -> None:
        r"""Apply minimal distortion principle to estimated spectrograms."""
        assert self.scale_restoration, "Set self.scale_restoration=True."

        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)
        Y_scaled = minimal_distortion_principle(Y, reference=X, reference_id=self.reference_id)
        X = X.transpose(1, 0, 2)
        Y = Y_scaled.transpose(1, 0, 2)
        X_Hermite = X.transpose(0, 2, 1).conj()
        W_scaled = Y @ X_Hermite @ np.linalg.inv(X @ X_Hermite)

        self.output, self.demix_filter = Y_scaled, W_scaled


class BlockDecompositionIPSDTABase(IPSDTABase):
    r"""Base class of independent positive semidefinite tensor analysis (IPSDTA) \
    using block decomposition of bases.

    Args:
        n_basis (int):
            Number of PSDTF bases.
        n_blocks (int):
            Number of sub-blocks.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        n_blocks: int,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[
                Callable[["BlockDecompositionIPSDTABase"], None],
                List[Callable[["BlockDecompositionIPSDTABase"], None]],
            ]
        ] = None,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis=n_basis,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
            rng=rng,
        )

        self.n_blocks = n_blocks

    def __repr__(self) -> str:
        s = "IPSDTA("
        s += "n_basis={n_basis}"
        s += ", n_blocks={n_blocks}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        **kwargs,
    ) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            kwargs:
                Keyword arguments to set as attributes of IPSDTA.
        """
        assert self.input is not None, "Specify data!"

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        n_channels, n_bins, n_frames = X.shape
        n_sources = n_channels  # n_channels == n_sources

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        if not hasattr(self, "demix_filter"):
            W = np.eye(n_sources, n_channels, dtype=np.complex128)
            W = np.tile(W, reps=(n_bins, 1, 1))
        else:
            if self.demix_filter is None:
                W = None
            else:
                # To avoid overwriting ``demix_filter`` given by keyword arguments.
                W = self.demix_filter.copy()

        self.demix_filter = W
        self.output = self.separate(X, demix_filter=W)

        self._init_block_decomposition_psdtf(flooring_fn=flooring_fn, rng=self.rng)

    def _init_block_decomposition_psdtf(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        r"""Initialize PSDTF using block decomposition of bases.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        n_basis = self.n_basis
        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames
        n_blocks = self.n_blocks
        n_remains = self.n_remains

        n_neighbors = n_bins // n_blocks

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if rng is None:
            rng = np.random.default_rng()

        if not hasattr(self, "basis"):
            # should be positive semi-definite
            eye = np.eye(n_neighbors, dtype=np.complex128)
            rand = rng.random((n_sources, n_basis, n_blocks - n_remains, n_neighbors))
            T = rand[..., np.newaxis] * eye

            if n_remains > 0:
                eye = np.eye(n_neighbors + 1, dtype=np.complex128)
                rand = rng.random((n_sources, n_basis, n_remains, n_neighbors + 1))
                T_high = rand[..., np.newaxis] * eye

                T = T, T_high
        else:
            # To avoid overwriting.
            if n_remains > 0:
                T_low, T_high = self.basis
                T = T_low.copy(), T_high.copy()
            else:
                T = self.basis.copy()

        if not hasattr(self, "activation"):
            V = rng.random((n_sources, n_basis, n_frames))
            V = flooring_fn(V)
        else:
            # To avoid overwriting.
            V = self.activation.copy()

        self.basis, self.activation = T, V

        if self.source_normalization:
            self.normalize_block_decomposition_psdtf()

    @property
    def n_remains(self) -> int:
        if not hasattr(self, "n_bins"):
            raise AttributeError("Since n_bins is not defined, n_remains cannot be computed.")

        return self.n_bins % self.n_blocks

    def reconstruct_block_decomposition_psdtf(
        self, basis: np.ndarray, activation: np.ndarray, axis1: int = -2, axis2: int = -1
    ) -> np.ndarray:
        r"""Reconstruct PSDTF using block decomposition of bases.

        Args:
            basis (numpy.ndarray):
                Basis matrix.
                The shape is (n_sources, n_basis, n_blocks, n_neighbors, n_neighbors)
                if ``axis1=-1`` and ``axis2=-2``.
                Otherwise, (n_sources, n_blocks, n_neighbors, n_neighbors, n_basis).
            activation (numpy.ndarray):
                Activation matrix.
                The shape is (n_sources, n_basis, n_frames).
            axis1 (int):
                First axis of covariance matrix. Default: ``-2``.
            axis2 (int):
                Second axis of covariance matrix. Default: ``-1``.

        Returns:
            numpy.ndarray of reconstructed PSDTF.
            The shape is (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).
        """

        def _reconstruct(
            basis: np.ndarray, activation: np.ndarray, axis1: int = -2, axis2: int = -1
        ) -> np.ndarray:
            r"""Reconstruct PSDTF using block decomposition of bases.

            Args:
                basis (numpy.ndarray):
                    Basis matrix.
                    The shape is (n_sources, n_basis, n_blocks, n_neighbors, n_neighbors)
                    if ``axis1=-1`` and ``axis2=-2``.
                    Otherwise, (n_sources, n_blocks, n_neighbors, n_neighbors, n_basis).
                activation (numpy.ndarray):
                    Activation matrix.
                    The shape is (n_sources, n_basis, n_frames).
                axis1 (int):
                    First axis of covariance matrix. Default: ``-2``.
                axis2 (int):
                    Second axis of covariance matrix. Default: ``-1``.

            Returns:
                numpy.ndarray of reconstructed PSDTF.
                The shape is (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).
            """
            na = np.newaxis
            T, V = basis, activation
            n_dims = T.ndim

            axis1 = n_dims + axis1 if axis1 < 0 else axis1
            axis2 = n_dims + axis2 if axis2 < 0 else axis2

            assert (axis1 == 2 and axis2 == 3) or (axis1 == 3 and axis2 == 4)

            if axis1 == 2 and axis2 == 3:
                T = T.transpose(0, 4, 1, 2, 3)

            R = np.sum(
                T[:, :, na, :, :, :] * V[:, :, :, na, na, na],
                axis=1,
            )
            R = to_psd(R, axis1=3, axis2=4)

            return R

        if type(basis) is tuple:
            assert self.n_remains > 0, "n_remains is expected to be positive."

            T_low, T_high = basis
            V = activation
            R_low = _reconstruct(T_low, V, axis1=axis1, axis2=axis2)
            R_high = _reconstruct(T_high, V, axis1=axis1, axis2=axis2)
            R = R_low, R_high
        else:
            T = basis
            V = activation
            R = _reconstruct(T, V, axis1=axis1, axis2=axis2)

        return R

    def normalize_block_decomposition_psdtf(self, axis1: int = -2, axis2: int = -1) -> None:
        r"""Normalize PSDTF parameters using block decomposition of bases.

        Args:
            axis1 (int):
                First axis of covariance matrix. Default: ``-2``.
            axis2 (int):
                Second axis of covariance matrix. Default: ``-1``.
        """
        source_normalization = self.source_normalization
        n_remains = self.n_remains
        na = np.newaxis
        T, V = self.basis, self.activation

        assert source_normalization, "Set source_normalization."

        if n_remains > 0:
            T_low, T_high = T
            trace_low = np.trace(T_low, axis1=axis1, axis2=axis2).real
            trace_high = np.trace(T_high, axis1=axis1, axis2=axis2).real
            trace = np.sum(trace_low, axis=-1) + np.sum(trace_high, axis=-1)
            T_low = T_low / trace[:, :, na, na, na]
            T_high = T_high / trace[:, :, na, na, na]
            T = T_low, T_high
        else:
            trace = np.trace(T, axis1=axis1, axis2=axis2).real
            trace = np.sum(trace, axis=-1)
            T = T / trace[:, :, na, na, na]

        V = V * trace[:, :, na]

        self.basis, self.activation = T, V


class GaussIPSDTA(BlockDecompositionIPSDTABase):
    r"""Independent positive semidefinite tensor analysis (IPSDTA) \
    on Gaussian distribution.

    Args:
        n_basis (int):
            Number of PSDTF bases.
        n_blocks (int):
            Number of sub-blocks.
        source_algorithm (str):
            Algorithm for PSDTF updates.
            Choose ``EM``, or ``MM``. Default: ``MM``.
        spatial_algorithm (str):
            Algorithm for demixing filter updates.
            Choose ``FPI``, or ``VCD``. Default: ``VCD``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        source_normalization (bool):
            If ``source_normalization=True``, normalize PSDTF parameters.
            Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        n_blocks: int,
        source_algorithm: str = "MM",
        spatial_algorithm: str = "VCD",
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[
                Callable[["GaussIPSDTA"], None],
                List[Callable[["GaussIPSDTA"], None]],
            ]
        ] = None,
        source_normalization: Optional[Union[bool, str]] = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis,
            n_blocks,
            flooring_fn,
            callbacks,
            scale_restoration,
            record_loss,
            reference_id,
            rng,
        )

        assert source_algorithm in source_algorithms, "Not support {}.".format(source_algorithms)
        assert spatial_algorithm in spatial_algorithms, "Not support {}.".format(spatial_algorithms)

        self.source_algorithm = source_algorithm
        self.spatial_algorithm = spatial_algorithm
        self.source_normalization = source_normalization

    def __repr__(self) -> str:
        s = "GaussIPSDTA("
        s += "n_basis={n_basis}"
        s += ", n_blocks={n_blocks}"
        s += ", source_algorithm={source_algorithm}"
        s += ", spatial_algorithm={spatial_algorithm}"
        s += ", source_normalization={source_normalization}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of IPSDTA.
        """
        super()._reset(**kwargs)

        if self.spatial_algorithm == "FPI":
            if not hasattr(self, "fixed_point"):
                n_sources = self.n_sources
                n_bins = self.n_bins

                self.fixed_point = np.ones((n_sources, n_bins), dtype=np.complex128)
            else:
                self.fixed_point = self.fixed_point.copy()

            raise NotImplementedError("IPSDTA with fixed-point iteration is not supported.")

    def update_once(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF parameters and demixing filters once.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_source_model(flooring_fn=flooring_fn)
        self.update_spatial_model(flooring_fn=flooring_fn)

    def update_source_model(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF basis matrices and activations.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if self.source_algorithm == "MM":
            self.update_source_model_mm(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Not support {}.".format(self.source_algorithm))

        if self.source_normalization:
            self.normalize_block_decomposition_psdtf()

    def update_source_model_mm(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF basis matrices and activations by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_basis_mm(flooring_fn=flooring_fn)
        self.update_activation_mm()

    def update_basis_mm(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF basis matrices by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        n_sources = self.n_sources
        n_frames = self.n_frames

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _update_basis_mm(
            basis: np.ndarray, activation: np.ndarray, separated: np.ndarray = None
        ) -> np.ndarray:
            r"""
            Args:
                basis: (n_sources, n_basis, n_blocks, n_neighbors, n_neighbors)
                activation: (n_sources, n_basis, n_frames)
                separated: (n_sources, n_blocks, n_neighbors, n_frames)

            Returns:
                numpy.ndarray of updated basis matrix.
            """
            T, V = basis, activation
            Y = separated
            na = np.newaxis

            R = self.reconstruct_block_decomposition_psdtf(T, V)
            R_inverse = np.linalg.inv(R)
            Y = Y.transpose(0, 3, 1, 2)
            YY_Hermite = Y[:, :, :, :, na] @ Y[:, :, :, na, :].conj()
            RYYR = R_inverse @ YY_Hermite @ R_inverse

            P = np.mean(
                V[:, :, :, na, na, na] * R_inverse[:, na, :, :, :, :],
                axis=2,
            )
            Q = np.mean(
                V[:, :, :, na, na, na] * RYYR[:, na, :, :, :, :],
                axis=2,
            )
            TQT = T @ Q @ T

            P = to_psd(P, flooring_fn=flooring_fn)
            TQT = to_psd(TQT, flooring_fn=flooring_fn)

            # geometric mean of P^(-1) and TQT
            T = gmeanmh(P, TQT, type=2)
            T = to_psd(T, flooring_fn=flooring_fn)

            return T

        n_bins = self.n_bins
        n_blocks = self.n_blocks
        n_remains = self.n_remains
        n_neighbors = n_bins // n_blocks

        X, W = self.input, self.demix_filter
        T, V = self.basis, self.activation
        Y = self.separate(X, demix_filter=W)

        if n_remains > 0:
            T_low, T_high = T
            Y_low, Y_high = np.split(Y, [(n_blocks - n_remains) * n_neighbors], axis=1)
            Y_low = Y_low.reshape(n_sources, n_blocks - n_remains, n_neighbors, n_frames)
            Y_high = Y_high.reshape(n_sources, n_remains, n_neighbors + 1, n_frames)

            T_low = _update_basis_mm(T_low, V, separated=Y_low)
            T_high = _update_basis_mm(T_high, V, separated=Y_high)
            T = T_low, T_high
        else:
            Y = Y.reshape(n_sources, n_blocks, n_neighbors, n_frames)
            T = _update_basis_mm(T, V, separated=Y)

        self.basis = T

    def update_activation_mm(self) -> None:
        r"""Update PSDTF activations by MM algorithm."""

        def _compute_traces(
            basis: np.ndarray, activation: np.ndarray, separated: np.ndarray = None
        ) -> Tuple[np.ndarray, np.ndarray]:
            r"""
            Args:
                basis: (n_sources, n_basis, n_blocks, n_neighbors, n_neighbors)
                activation: (n_sources, n_basis, n_frames)
                separated: (n_sources, n_blocks, n_neighbors, n_frames)

            Returns:
                Tuple of numerator and denominator.
                Type of each item is ``numpy.ndarray``.
            """
            T, V = basis, activation
            Y = separated
            na = np.newaxis

            R = self.reconstruct_block_decomposition_psdtf(T, V)
            R_inverse = np.linalg.inv(R)
            Y = Y.transpose(0, 3, 1, 2)
            YY_Hermite = Y[:, :, :, :, na] @ Y[:, :, :, na, :].conj()
            RYYR = R_inverse @ YY_Hermite @ R_inverse

            num = np.trace(RYYR[:, na, :] @ T[:, :, na], axis1=-2, axis2=-1)
            denom = np.trace(R_inverse[:, na, :] @ T[:, :, na], axis1=-2, axis2=-1)
            num = np.real(num).sum(axis=-1)
            denom = np.real(denom).sum(axis=-1)

            return num, denom

        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames
        n_blocks = self.n_blocks
        n_remains = self.n_remains
        n_neighbors = n_bins // n_blocks

        X, W = self.input, self.demix_filter
        T, V = self.basis, self.activation
        Y = self.separate(X, demix_filter=W)

        if n_remains > 0:
            T_low, T_high = T
            Y_low, Y_high = np.split(Y, [(n_blocks - n_remains) * n_neighbors], axis=1)
            Y_low = Y_low.reshape(n_sources, n_blocks - n_remains, n_neighbors, n_frames)
            Y_high = Y_high.reshape(n_sources, n_remains, n_neighbors + 1, n_frames)

            num_low, denom_low = _compute_traces(T_low, V, separated=Y_low)
            num_high, denom_high = _compute_traces(T_high, V, separated=Y_high)

            num = num_low + num_high
            denom = denom_low + denom_high
        else:
            Y = Y.reshape(n_sources, n_blocks, n_neighbors, n_frames)
            num, denom = _compute_traces(T, V, separated=Y)

        self.activation = V * np.sqrt(num / denom)

    def update_spatial_model(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if self.spatial_algorithm == "VCD":
            self.update_spatial_model_vcd(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Not support {}.".format(self.spatial_algorithm))

    def update_spatial_model_vcd(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once by VCD.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _update(input: np.ndarray, demix_filter: np.ndarray, covariance: np.ndarray):
            r"""
            Args:
                input (np.ndarray):
                    Mixture spectrogram.
                    The shape is (n_channels, n_blocks, n_neighbors, n_frames).
                demix_filter (np.ndarray):
                    Demixing filter split by frequnecy bands.
                    The shape is (n_blocks, n_neighbors, n_sources, n_channels).
                covariance (np.ndarray):
                    Rconstructed PSDTF.
                    The shape is (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).

            Returns:
                np.ndarray of demixing filters after update.
            """
            X, W = input, demix_filter
            R = covariance

            XX = X[:, na, :, :, na] * X[na, :, :, na, :].conj()
            XX = XX.transpose(2, 3, 4, 0, 1, 5)

            R_inverse = np.linalg.inv(R)
            R_inverse = R_inverse.transpose(2, 4, 3, 0, 1)

            RXX = np.mean(R_inverse[:, :, :, :, na, na] * XX[:, :, :, na, :, :], axis=-1)

            W = update_by_block_decomposition_vcd(
                W, weighted_covariance=RXX, singular_fn=lambda x: np.abs(x) < flooring_fn(0)
            )

            return W

        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins, n_frames = self.n_bins, self.n_frames
        n_blocks = self.n_blocks
        n_remains = self.n_remains
        na = np.newaxis

        n_neighbors = n_bins // n_blocks

        X, W = self.input, self.demix_filter
        T, V = self.basis, self.activation

        R = self.reconstruct_block_decomposition_psdtf(T, V)

        if n_remains > 0:
            X_low, X_high = np.split(X, [(n_blocks - n_remains) * n_neighbors], axis=1)
            W_low, W_high = np.split(W, [(n_blocks - n_remains) * n_neighbors], axis=0)
            R_low, R_high = R

            # Lower frequency
            X_low = X_low.reshape(n_channels, n_blocks - n_remains, n_neighbors, n_frames)
            W_low = W_low.reshape(n_blocks - n_remains, n_neighbors, n_sources, n_channels)
            W_low = _update(X_low, demix_filter=W_low, covariance=R_low)

            # Higher frequency
            X_high = X_high.reshape(n_channels, n_remains, n_neighbors + 1, n_frames)
            W_high = W_high.reshape(n_remains, n_neighbors + 1, n_sources, n_channels)
            W_high = _update(X_high, demix_filter=W_high, covariance=R_high)

            W_low = W_low.reshape((n_blocks - n_remains) * n_neighbors, n_sources, n_channels)
            W_high = W_high.reshape(n_remains * (n_neighbors + 1), n_sources, n_channels)
            W = np.concatenate([W_low, W_high], axis=0)
        else:
            X = X.reshape(n_channels, n_blocks, n_neighbors, n_frames)
            W = W.reshape(n_blocks, n_neighbors, n_sources, n_channels)
            W = _update(X, demix_filter=W, covariance=R)
            W = W.reshape(n_blocks * n_neighbors, n_sources, n_channels)

        self.demix_filter = W

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        Returns:
            Computed loss.
        """

        def _compute_block_decomposition_loss(
            separated: np.ndarray, demix_filter: np.ndarray, covariance: np.ndarray
        ) -> float:
            r"""
            Args:
                separated (np.ndarray):
                    Separated signal with shape of (n_sources, n_frames, n_blocks, n_neighbors).
                demix_filter (np.ndarray):
                    Demixing filters with shape of (n_blocks, n_neighbors, n_sources, n_channels).
                covariance:
                    Covariance matrix with shape of
                    (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).
            """
            Y, W = separated, demix_filter
            R = covariance

            n_sources, n_frames, n_blocks, n_neighbors = Y.shape

            Y = Y.reshape(n_sources, n_frames, n_blocks, n_neighbors, 1)
            R_inverse = np.linalg.inv(R)
            Y_Hermite = np.swapaxes(Y, 3, 4).conj()
            YRY = np.sum(Y_Hermite @ R_inverse @ Y, axis=(0, 2, 3, 4))
            YRY = np.real(YRY)
            YRY = np.maximum(YRY, 0)
            _, logdetR = np.linalg.slogdet(R)
            logdetR = logdetR.sum(axis=(0, 2))
            logdetW = self.compute_logdet(W)

            loss = np.mean(YRY + logdetR, axis=0) - 2 * logdetW.sum(axis=(0, 1))
            loss = loss.item()

            return loss

        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins, n_frames = self.n_bins, self.n_frames
        n_blocks = self.n_blocks
        n_remains = self.n_remains

        n_neighbors = n_bins // n_blocks

        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)
        Y = Y.transpose(0, 2, 1)
        T, V = self.basis, self.activation

        R = self.reconstruct_block_decomposition_psdtf(T, V)

        if n_remains > 0:
            Y_low, Y_high = np.split(Y, [(n_blocks - n_remains) * n_neighbors], axis=2)
            W_low, W_high = np.split(W, [(n_blocks - n_remains) * n_neighbors], axis=0)
            R_low, R_high = R

            Y_low = Y_low.reshape(n_sources, n_frames, (n_blocks - n_remains), n_neighbors)
            Y_high = Y_high.reshape(n_sources, n_frames, n_remains, n_neighbors + 1)
            W_low = W_low.reshape((n_blocks - n_remains), n_neighbors, n_sources, n_channels)
            W_high = W_high.reshape(n_remains, n_neighbors + 1, n_sources, n_channels)

            loss_low = _compute_block_decomposition_loss(
                Y_low, demix_filter=W_low, covariance=R_low
            )
            loss_high = _compute_block_decomposition_loss(
                Y_high, demix_filter=W_high, covariance=R_high
            )

            loss = loss_low + loss_high
        else:
            Y = Y.reshape(n_sources, n_frames, n_blocks, n_neighbors)
            W = W.reshape(n_blocks, n_neighbors, n_sources, n_channels)

            loss = _compute_block_decomposition_loss(Y, demix_filter=W, covariance=R)

        return loss


class TIPSDTA(BlockDecompositionIPSDTABase):
    r"""Independent positive semidefinite tensor analysis (IPSDTA) \
    on Student's t distribution.

    Args:
        n_basis (int):
            Number of PSDTF bases.
        n_blocks (int):
            Number of sub-blocks.
        dof (float):
            Degree of freedom parameter.
        source_algorithm (str):
            Algorithm for PSDTF updates.
            Only ``MM`` is supported. Default: ``MM``.
        spatial_algorithm (str):
            Algorithm for demixing filter updates.
            Only ``VCD`` is supported. Default: ``VCD``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        source_normalization (bool):
            If ``source_normalization=True``, normalize PSDTF parameters.
            Default: ``True``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        n_blocks: int,
        dof: float,
        source_algorithm: str = "MM",
        spatial_algorithm: str = "VCD",
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[
                Callable[["GaussIPSDTA"], None],
                List[Callable[["GaussIPSDTA"], None]],
            ]
        ] = None,
        source_normalization: Optional[Union[bool, str]] = True,
        scale_restoration: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis,
            n_blocks,
            flooring_fn,
            callbacks,
            scale_restoration,
            record_loss,
            reference_id,
            rng,
        )

        assert source_algorithm in source_algorithms, "Not support {}.".format(source_algorithm)
        assert spatial_algorithm in spatial_algorithms, "Not support {}.".format(spatial_algorithm)

        self.dof = dof
        self.source_algorithm = source_algorithm
        self.source_normalization = source_normalization
        self.spatial_algorithm = spatial_algorithm

    def __repr__(self) -> str:
        s = "TIPSDTA("
        s += "n_basis={n_basis}"
        s += ", n_blocks={n_blocks}"
        s += ", dof={dof}"
        s += ", source_algorithm={source_algorithm}"
        s += ", spatial_algorithm={spatial_algorithm}"
        s += ", source_normalization={source_normalization}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def update_once(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF parameters and demixing filters once.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_source_model(flooring_fn=flooring_fn)
        self.update_spatial_model(flooring_fn=flooring_fn)

    def update_source_model(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF basis matrices and activations.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if self.source_algorithm == "MM":
            self.update_source_model_mm(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Not support {}.".format(self.source_algorithm))

        if self.source_normalization:
            self.normalize_block_decomposition_psdtf()

    def update_source_model_mm(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF basis matrices and activations by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_basis_mm(flooring_fn=flooring_fn)
        self.update_activation_mm()

    def update_basis_mm(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update PSDTF basis matrices by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        n_sources = self.n_sources
        n_frames = self.n_frames

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _quadratic(Y: np.ndarray, R: np.ndarray) -> np.ndarray:
            r"""
            Args:
                Y (np.ndarray):
                    Separated spectrams with shape of
                    (n_sources, n_blocks, n_neighbors, n_frames).
                R (np.ndarray):
                    Covariance matrix with shape of
                    (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).

            Returns:
                Quadratic forms with shape of (n_sources, n_frames).
            """
            Y = Y.transpose(0, 3, 1, 2)
            R_inverse = np.linalg.inv(R)

            YRY = quadratic(Y, R_inverse)
            YRY = np.real(YRY)
            YRY = np.maximum(YRY, 0)
            YRY = YRY.sum(axis=-1)

            return YRY

        def _update_basis_mm(
            basis: np.ndarray,
            activation: np.ndarray,
            separated: np.ndarray = None,
            weight: np.ndarray = None,
        ) -> np.ndarray:
            r"""
            Args:
                basis: (n_sources, n_basis, n_blocks, n_neighbors, n_neighbors)
                activation: (n_sources, n_basis, n_frames)
                separated: (n_sources, n_blocks, n_neighbors, n_frames)
                weight: (n_sources, n_frames)

            Returns:
                numpy.ndarray of updated basis matrix.
            """
            T, V = basis, activation
            Y = separated
            pi = weight
            na = np.newaxis

            R = self.reconstruct_block_decomposition_psdtf(T, V)
            R_inverse = np.linalg.inv(R)
            Y = Y.transpose(0, 3, 1, 2)
            YY_Hermite = Y[:, :, :, :, na] @ Y[:, :, :, na, :].conj()
            RYYR = R_inverse @ YY_Hermite @ R_inverse
            piRYYR = pi[:, :, na, na, na] * RYYR

            P = np.mean(
                V[:, :, :, na, na, na] * R_inverse[:, na, :, :, :, :],
                axis=2,
            )
            Q = np.mean(
                V[:, :, :, na, na, na] * piRYYR[:, na, :, :, :, :],
                axis=2,
            )
            Q = to_psd(Q, flooring_fn=flooring_fn)
            Q_sqrt = sqrtmh(Q)

            QTPTQ = Q_sqrt @ T @ P @ T @ Q_sqrt
            QTPTQ = to_psd(QTPTQ, flooring_fn=flooring_fn)
            T = T @ Q_sqrt @ invsqrtmh(QTPTQ, flooring_fn=flooring_fn) @ Q_sqrt @ T
            T = to_psd(T, flooring_fn=flooring_fn)

            return T

        n_bins = self.n_bins
        n_blocks = self.n_blocks
        n_remains = self.n_remains
        n_neighbors = n_bins // n_blocks

        nu = self.dof

        X, W = self.input, self.demix_filter
        T, V = self.basis, self.activation

        Y = self.separate(X, demix_filter=W)
        R = self.reconstruct_block_decomposition_psdtf(T, V)

        if n_remains > 0:
            T_low, T_high = T
            Y_low, Y_high = np.split(Y, [(n_blocks - n_remains) * n_neighbors], axis=1)
            Y_low = Y_low.reshape(n_sources, n_blocks - n_remains, n_neighbors, n_frames)
            Y_high = Y_high.reshape(n_sources, n_remains, n_neighbors + 1, n_frames)
            R_low, R_high = R

            YRY_low = _quadratic(Y_low, R_low)
            YRY_high = _quadratic(Y_high, R_high)

            YRY = YRY_low + YRY_high
            pi = (nu + 2 * n_bins) / (nu + 2 * YRY)

            T_low = _update_basis_mm(T_low, V, separated=Y_low, weight=pi)
            T_high = _update_basis_mm(T_high, V, separated=Y_high, weight=pi)
            T = T_low, T_high
        else:
            Y = Y.reshape(n_sources, n_blocks, n_neighbors, n_frames)
            YRY = _quadratic(Y, R)
            pi = (nu + 2 * n_bins) / (nu + 2 * YRY)

            T = _update_basis_mm(T, V, separated=Y, weight=pi)

        self.basis = T

    def update_activation_mm(self) -> None:
        r"""Update PSDTF activations by MM algorithm."""

        def _quadratic(Y: np.ndarray, R: np.ndarray) -> np.ndarray:
            r"""
            Args:
                Y (np.ndarray):
                    Separated spectrams with shape of
                    (n_sources, n_blocks, n_neighbors, n_frames).
                R (np.ndarray):
                    Covariance matrix with shape of
                    (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).

            Returns:
                Quadratic forms with shape of (n_sources, n_frames).
            """
            Y = Y.transpose(0, 3, 1, 2)
            R_inverse = np.linalg.inv(R)

            YRY = quadratic(Y, R_inverse)
            YRY = np.real(YRY)
            YRY = np.maximum(YRY, 0)
            YRY = YRY.sum(axis=-1)

            return YRY

        def _compute_traces(
            basis: np.ndarray,
            activation: np.ndarray,
            separated: np.ndarray = None,
            weight: np.ndarray = None,
        ) -> Tuple[np.ndarray, np.ndarray]:
            r"""
            Args:
                basis: (n_sources, n_basis, n_blocks, n_neighbors, n_neighbors)
                activation: (n_sources, n_basis, n_frames)
                separated: (n_sources, n_blocks, n_neighbors, n_frames)

            Returns:
                Tuple of numerator and denominator.
                Type of each item is ``numpy.ndarray``.
            """
            T, V = basis, activation
            Y = separated.transpose(0, 3, 1, 2)
            pi = weight
            na = np.newaxis

            R = self.reconstruct_block_decomposition_psdtf(T, V)
            R_inverse = np.linalg.inv(R)
            YY_Hermite = Y[:, :, :, :, na] @ Y[:, :, :, na, :].conj()
            RYYR = R_inverse @ YY_Hermite @ R_inverse

            piRYYR = pi[:, :, na, na, na] * RYYR

            num = np.trace(piRYYR[:, na, :] @ T[:, :, na], axis1=-2, axis2=-1)
            denom = np.trace(R_inverse[:, na, :] @ T[:, :, na], axis1=-2, axis2=-1)
            num = np.real(num).sum(axis=-1)
            denom = np.real(denom).sum(axis=-1)

            return num, denom

        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames
        n_blocks = self.n_blocks
        n_remains = self.n_remains
        n_neighbors = n_bins // n_blocks

        nu = self.dof

        X, W = self.input, self.demix_filter
        T, V = self.basis, self.activation

        Y = self.separate(X, demix_filter=W)
        R = self.reconstruct_block_decomposition_psdtf(T, V)

        if n_remains > 0:
            T_low, T_high = T
            Y_low, Y_high = np.split(Y, [(n_blocks - n_remains) * n_neighbors], axis=1)
            Y_low = Y_low.reshape(n_sources, n_blocks - n_remains, n_neighbors, n_frames)
            Y_high = Y_high.reshape(n_sources, n_remains, n_neighbors + 1, n_frames)
            R_low, R_high = R

            YRY_low = _quadratic(Y_low, R_low)
            YRY_high = _quadratic(Y_high, R_high)

            YRY = YRY_low + YRY_high
            pi = (nu + 2 * n_bins) / (nu + 2 * YRY)

            num_low, denom_low = _compute_traces(T_low, V, separated=Y_low, weight=pi)
            num_high, denom_high = _compute_traces(T_high, V, separated=Y_high, weight=pi)

            num = num_low + num_high
            denom = denom_low + denom_high
        else:
            Y = Y.reshape(n_sources, n_blocks, n_neighbors, n_frames)
            YRY = _quadratic(Y, R)
            pi = (nu + 2 * n_bins) / (nu + 2 * YRY)

            num, denom = _compute_traces(T, V, separated=Y, weight=pi)

        self.activation = V * np.sqrt(num / denom)

    def update_spatial_model(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if self.spatial_algorithm == "VCD":
            self.update_spatial_model_vcd(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Not support {}.".format(self.spatial_algorithm))

    def update_spatial_model_vcd(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update demixing filters once by VCD.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _quadratic(Y: np.ndarray, R: np.ndarray) -> np.ndarray:
            r"""
            Args:
                Y (np.ndarray):
                    Separated spectrams with shape of
                    (n_sources, n_blocks, n_neighbors, n_frames).
                R (np.ndarray):
                    Covariance matrix with shape of
                    (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).

            Returns:
                Quadratic forms with shape of (n_sources, n_frames).
            """
            Y = Y.transpose(2, 3, 0, 1)
            R_inverse = np.linalg.inv(R)

            YRY = quadratic(Y, R_inverse)
            YRY = np.real(YRY)
            YRY = np.maximum(YRY, 0)
            YRY = YRY.sum(axis=-1)

            return YRY

        def _update(
            input: np.ndarray,
            demix_filter: np.ndarray,
            covariance: np.ndarray,
            weight: np.ndarray = None,
        ):
            X, W = input, demix_filter
            R = covariance
            pi = weight

            na = np.newaxis

            XX = X[:, na, :, :, na] * X[na, :, :, na, :].conj()
            XX = XX.transpose(2, 3, 4, 0, 1, 5)

            R_inverse = np.linalg.inv(R)
            R_inverse = R_inverse.transpose(2, 4, 3, 0, 1)
            pi_R_inverse = pi * R_inverse

            RXX = np.mean(pi_R_inverse[:, :, :, :, na, na] * XX[:, :, :, na, :, :], axis=-1)

            W = update_by_block_decomposition_vcd(
                W, weighted_covariance=RXX, singular_fn=lambda x: np.abs(x) < flooring_fn(0)
            )

            return W

        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins, n_frames = self.n_bins, self.n_frames
        n_blocks = self.n_blocks
        n_remains = self.n_remains
        n_neighbors = n_bins // n_blocks

        nu = self.dof

        X, W = self.input, self.demix_filter
        T, V = self.basis, self.activation

        R = self.reconstruct_block_decomposition_psdtf(T, V)

        if n_remains > 0:
            X_low, X_high = np.split(X, [(n_blocks - n_remains) * n_neighbors], axis=1)
            W_low, W_high = np.split(W, [(n_blocks - n_remains) * n_neighbors], axis=0)
            R_low, R_high = R

            # Lower frequency
            X_low = X_low.reshape(n_channels, n_blocks - n_remains, n_neighbors, n_frames)
            W_low = W_low.reshape(n_blocks - n_remains, n_neighbors, n_sources, n_channels)
            Y_low = W_low @ X_low.transpose(1, 2, 0, 3)

            # Higher frequency
            X_high = X_high.reshape(n_channels, n_remains, n_neighbors + 1, n_frames)
            W_high = W_high.reshape(n_remains, n_neighbors + 1, n_sources, n_channels)
            Y_high = W_high @ X_high.transpose(1, 2, 0, 3)

            YRY_low = _quadratic(Y_low, R_low)
            YRY_high = _quadratic(Y_high, R_high)

            YRY = YRY_low + YRY_high
            pi = (nu + 2 * n_bins) / (nu + 2 * YRY)

            W_low = _update(X_low, demix_filter=W_low, covariance=R_low, weight=pi)
            W_high = _update(X_high, demix_filter=W_high, covariance=R_high, weight=pi)

            W_low = W_low.reshape((n_blocks - n_remains) * n_neighbors, n_sources, n_channels)
            W_high = W_high.reshape(n_remains * (n_neighbors + 1), n_sources, n_channels)
            W = np.concatenate([W_low, W_high], axis=0)
        else:
            X = X.reshape(n_channels, n_blocks, n_neighbors, n_frames)
            W = W.reshape(n_blocks, n_neighbors, n_sources, n_channels)
            Y = W @ X.transpose(1, 2, 0, 3)

            YRY = _quadratic(Y, R)
            pi = (nu + 2 * n_bins) / (nu + 2 * YRY)

            W = _update(X, demix_filter=W, covariance=R, weight=pi)
            W = W.reshape(n_blocks * n_neighbors, n_sources, n_channels)

        self.demix_filter = W

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        Returns:
            Computed loss.
        """

        def _quadratic(Y: np.ndarray, R: np.ndarray) -> np.ndarray:
            r"""
            Args:
                Y (np.ndarray):
                    Separated spectrams with shape of
                    (n_sources, n_blocks, n_neighbors, n_frames).
                R (np.ndarray):
                    Covariance matrix with shape of
                    (n_sources, n_frames, n_blocks, n_neighbors, n_neighbors).

            Returns:
                Quadratic forms with shape of (n_sources, n_frames).
            """
            Y = Y.transpose(0, 3, 1, 2)
            R_inverse = np.linalg.inv(R)

            YRY = quadratic(Y, R_inverse)
            YRY = np.real(YRY)
            YRY = np.maximum(YRY, 0)
            YRY = YRY.sum(axis=-1)

            return YRY

        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins, n_frames = self.n_bins, self.n_frames

        nu = self.dof

        n_blocks = self.n_blocks
        n_remains = self.n_remains

        n_neighbors = n_bins // n_blocks

        X, W = self.input, self.demix_filter
        Y = self.separate(X, demix_filter=W)
        T, V = self.basis, self.activation

        R = self.reconstruct_block_decomposition_psdtf(T, V)

        if n_remains > 0:
            Y_low, Y_high = np.split(Y, [(n_blocks - n_remains) * n_neighbors], axis=1)
            W_low, W_high = np.split(W, [(n_blocks - n_remains) * n_neighbors], axis=0)
            R_low, R_high = R

            Y_low = Y_low.reshape(n_sources, (n_blocks - n_remains), n_neighbors, n_frames)
            Y_high = Y_high.reshape(n_sources, n_remains, n_neighbors + 1, n_frames)
            W_low = W_low.reshape((n_blocks - n_remains), n_neighbors, n_sources, n_channels)
            W_high = W_high.reshape(n_remains, n_neighbors + 1, n_sources, n_channels)

            YRY_low = _quadratic(Y_low, R_low)
            YRY_high = _quadratic(Y_high, R_high)

            YRY = YRY_low + YRY_high

            loss = np.sum(((nu + 2 * n_bins) / 2) * np.log(1 + (2 / nu) * YRY), axis=0)

            _, logdetR_low = np.linalg.slogdet(R_low)
            logdetR_low = logdetR_low.sum(axis=(0, 2))
            _, logdetR_high = np.linalg.slogdet(R_high)
            logdetR_high = logdetR_high.sum(axis=(0, 2))
            logdetR = logdetR_low + logdetR_high

            logdetW_low = self.compute_logdet(W_low)
            logdetW_high = self.compute_logdet(W_high)

            logdetW = logdetW_low.sum(axis=(0, 1)) + logdetW_high.sum(axis=(0, 1))
        else:
            Y = Y.reshape(n_sources, n_blocks, n_neighbors, n_frames)
            W = W.reshape(n_blocks, n_neighbors, n_sources, n_channels)

            YRY = _quadratic(Y, R)

            loss = np.sum(((nu + 2 * n_bins) / 2) * np.log(1 + (2 / nu) * YRY), axis=0)

            _, logdetR = np.linalg.slogdet(R)
            logdetR = logdetR.sum(axis=(0, 2))

            logdetW = self.compute_logdet(W)
            logdetW = logdetW.sum(axis=(0, 1))

        loss = np.mean(loss + logdetR, axis=0) - 2 * logdetW
        loss = loss.item()

        return loss



import os
import sys
import tempfile

import numpy as np
import pytest
from scipy.io import wavfile

from ssspy import wavread, wavwrite

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.io import save_invalid_wavfile  # noqa: E402

parameters_frame_offset = [0, 10]
parameters_num_frames = [None, 100]
parameters_channels_first = [True, False, None]
parameters_float = [True, False]
parameters_channels = [1, 2]


@pytest.mark.parametrize("frame_offset", parameters_frame_offset)
@pytest.mark.parametrize("num_frames", parameters_num_frames)
@pytest.mark.parametrize("channels_first", parameters_channels_first)
def test_wavread_monoral(frame_offset: int, num_frames: int, channels_first: bool):
    filename = "./tests/mock/audio/monoral_16k_5sec.wav"

    if channels_first is not None:
        return_2d = True
    else:
        return_2d = False

    # load file using scipy
    sample_rate_scipy, waveform_scipy = wavfile.read(filename)
    waveform_scipy = waveform_scipy / 2**15

    # load file using ssspy
    waveform_ssspy, sample_rate_ssspy = wavread(
        filename,
        frame_offset=frame_offset,
        num_frames=num_frames,
        return_2d=return_2d,
        channels_first=channels_first,
    )

    assert sample_rate_scipy == sample_rate_ssspy

    if return_2d:
        if channels_first:
            waveform_ssspy = waveform_ssspy.squeeze(axis=0)
        else:
            waveform_ssspy = waveform_ssspy.squeeze(axis=1)

    if num_frames is not None:
        assert np.all(waveform_scipy[frame_offset : frame_offset + num_frames] == waveform_ssspy)
    else:
        assert np.all(waveform_scipy[frame_offset:] == waveform_ssspy)


@pytest.mark.parametrize("frame_offset", parameters_frame_offset)
@pytest.mark.parametrize("num_frames", parameters_num_frames)
@pytest.mark.parametrize("channels_first", parameters_channels_first)
def test_wavread_stereo(frame_offset: int, num_frames: int, channels_first: bool):
    filename = "./tests/mock/audio/stereo_16k_5sec.wav"

    # load file using scipy
    sample_rate_scipy, waveform_scipy = wavfile.read(filename)
    waveform_scipy = waveform_scipy / 2**15

    # load file using ssspy
    waveform_ssspy, sample_rate_ssspy = wavread(
        filename, frame_offset=frame_offset, num_frames=num_frames, channels_first=channels_first
    )

    assert sample_rate_scipy == sample_rate_ssspy

    if channels_first:
        # same order as that of scipy
        waveform_ssspy = waveform_ssspy.transpose(1, 0)

    if num_frames is not None:
        assert np.all(waveform_scipy[frame_offset : frame_offset + num_frames] == waveform_ssspy)
    else:
        assert np.all(waveform_scipy[frame_offset:] == waveform_ssspy)


@pytest.mark.parametrize("frame_offset", parameters_frame_offset)
def test_wavread_invalid_monoral(frame_offset: int):
    filename = "./tests/mock/audio/monoral_16k_5sec.wav"
    max_frame = 5 * 16000
    valid_num_frames = max_frame - frame_offset

    # valid data
    wavread(filename, frame_offset=frame_offset, num_frames=valid_num_frames)

    # invalid memory size
    invalid_num_frames = valid_num_frames + 1

    with pytest.raises(ValueError) as e:
        wavread(filename, frame_offset=frame_offset, num_frames=invalid_num_frames)

    assert str(e.value) == f"num_frames={invalid_num_frames} exceeds maximum frame {max_frame}."


@pytest.mark.parametrize("frame_offset", parameters_frame_offset)
def test_wavread_invalid_stereo(frame_offset: int):
    filename = "./tests/mock/audio/stereo_16k_5sec.wav"
    max_frame = 5 * 16000
    valid_num_frames = max_frame - frame_offset

    # valid data
    wavread(filename, frame_offset=frame_offset, num_frames=valid_num_frames)

    # invalid memory size
    invalid_num_frames = valid_num_frames + 1

    with pytest.raises(ValueError) as e:
        wavread(filename, frame_offset=frame_offset, num_frames=invalid_num_frames)

    assert str(e.value) == f"num_frames={invalid_num_frames} exceeds maximum frame {max_frame}."


@pytest.mark.parametrize("is_float", parameters_float)
def test_wavio_1d(is_float: np.dtype):
    rng = np.random.default_rng(0)

    filename = "valid.wav"
    sample_rate = 16000
    duration = 5
    bits_per_sample = 16
    bytes_per_sample = bits_per_sample // 8
    num_frames = sample_rate * duration
    vmax = 2 ** (bits_per_sample - 1)

    waveform = rng.integers(-vmax, vmax, size=(num_frames,), dtype=f"<i{bytes_per_sample}")

    if is_float:
        waveform = waveform / vmax
        waveform_in = waveform.copy()
    else:
        waveform_in = waveform / vmax

    with tempfile.TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, filename)
        wavwrite(path, waveform, sample_rate=sample_rate)
        waveform_out, _ = wavread(path)

    assert np.all(waveform_in == waveform_out)


@pytest.mark.parametrize("is_float", parameters_float)
@pytest.mark.parametrize("n_channels", parameters_channels)
@pytest.mark.parametrize("channels_first", parameters_channels_first)
def test_wavio_2d(is_float: np.dtype, n_channels: int, channels_first: bool):
    rng = np.random.default_rng(0)

    filename = "valid.wav"
    sample_rate = 16000
    duration = 5
    bits_per_sample = 16
    bytes_per_sample = bits_per_sample // 8
    num_frames = sample_rate * duration
    vmax = 2 ** (bits_per_sample - 1)

    if channels_first:
        shape = (n_channels, num_frames)
    else:
        shape = (num_frames, n_channels)

    waveform = rng.integers(
        -vmax,
        vmax,
        size=shape,
        dtype=f"<i{bytes_per_sample}",
    )

    if is_float:
        waveform = waveform / vmax
        waveform_in = waveform.copy()
    else:
        waveform_in = waveform / vmax

    with tempfile.TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, filename)
        wavwrite(path, waveform, sample_rate=sample_rate, channels_first=channels_first)
        waveform_out, _ = wavread(path, return_2d=True, channels_first=channels_first)

    assert np.all(waveform_in == waveform_out)


def test_waveread_invalid_metadata():
    filename = "./tests/mock/audio/monoral_16k_5sec_invalid.wav"

    # invalid riff
    save_invalid_wavfile(filename, invalid_riff=True)

    with pytest.raises(NotImplementedError) as e:
        wavread(filename)

    assert str(e.value) == f"Not support {b'RIFX'}."

    # invalid ftype
    save_invalid_wavfile(filename, invalid_ftype=True)

    with pytest.raises(NotImplementedError) as e:
        wavread(filename)

    assert str(e.value) == f"Not support {b'wave'}."

    # invalid fmt chunk marker
    save_invalid_wavfile(filename, invalid_fmt_chunk_marker=True)

    with pytest.raises(NotImplementedError) as e:
        wavread(filename)

    assert str(e.value) == f"Not support {b'FMT '}."

    # invalid fmt chunk size
    save_invalid_wavfile(filename, invalid_fmt_chunk_size=True)

    with pytest.raises(NotImplementedError) as e:
        wavread(filename)

    assert str(e.value) == "Invalid header is detected."

    # invalid fmt
    save_invalid_wavfile(filename, invalid_fmt=True)

    with pytest.raises(NotImplementedError) as e:
        wavread(filename)

    assert str(e.value) == "Invalid header 0 is detected."

    # invalid fmt byte rate
    save_invalid_wavfile(filename, invalid_byte_rate=True)

    with pytest.raises(ValueError) as e:
        wavread(filename)

    assert str(e.value) == "Invalid header is detected."

    # invalid data chunk marker
    save_invalid_wavfile(filename, invalid_data_chunk_marker=True)

    with pytest.raises(NotImplementedError) as e:
        wavread(filename)

    assert str(e.value) == f"Not support {b'DATA'}."

    os.remove(filename)



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.mnmf import FastGaussMNMF, FastMNMFBase, GaussMNMF, MNMFBase

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.1
n_fft = 256
hop_length = 128
window = "hann"
n_bins = n_fft // 2 + 1
n_iter = 3
rng = np.random.default_rng(42)


parameters_diagonalizer_algorithm = ["IP", "IP1", "IP2"]
parameters_partitioning = [True, False]
parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_normalization = [True, False]
parameters_mnmf_base = [2]
parameters_mnmf = [
    (2, 2, 2, {}),
    (3, 2, 3, {}),
]


@pytest.mark.parametrize(
    "n_basis",
    parameters_mnmf_base,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_mnmf_base(
    n_basis: int,
    callbacks: Optional[Union[Callable[[MNMFBase], None], List[Callable[[MNMFBase], None]]]],
):
    ipsdta = MNMFBase(
        n_basis,
        callbacks=callbacks,
        record_loss=False,
        rng=rng,
    )

    print(ipsdta)


@pytest.mark.parametrize(
    "n_basis",
    parameters_mnmf_base,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_fast_mnmf_base(
    n_basis: int,
    callbacks: Optional[
        Union[Callable[[FastMNMFBase], None], List[Callable[[FastMNMFBase], None]]]
    ],
):
    ipsdta = FastMNMFBase(
        n_basis,
        callbacks=callbacks,
        record_loss=False,
        rng=rng,
    )

    print(ipsdta)


@pytest.mark.parametrize(
    "n_sources, n_channels, n_basis, reset_kwargs",
    parameters_mnmf,
)
@pytest.mark.parametrize(
    "partitioning",
    parameters_partitioning,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization)
def test_gauss_mnmf(
    n_sources: int,
    n_channels: int,
    n_basis: int,
    partitioning: bool,
    callbacks: Optional[Union[Callable[[GaussMNMF], None], List[Callable[[GaussMNMF], None]]]],
    normalization: Optional[Union[str, bool]],
    reset_kwargs: Dict[str, Any],
):
    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img[:n_channels], axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window=window, nperseg=n_fft, noverlap=n_fft - hop_length
    )

    mnmf = GaussMNMF(
        n_basis,
        n_sources=n_sources,
        partitioning=partitioning,
        callbacks=callbacks,
        normalization=normalization,
        rng=rng,
    )

    spectrogram_est = mnmf(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

    assert spectrogram_est.shape == (n_sources,) + spectrogram_mix.shape[1:]
    assert type(mnmf.loss[-1]) is float

    print(mnmf)


@pytest.mark.parametrize(
    "n_sources, n_channels, n_basis, reset_kwargs",
    parameters_mnmf,
)
@pytest.mark.parametrize("diagonalizer_algorithm", parameters_diagonalizer_algorithm)
@pytest.mark.parametrize(
    "partitioning",
    parameters_partitioning,
)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("normalization", parameters_normalization)
def test_fast_gauss_mnmf(
    n_sources: int,
    n_channels: int,
    n_basis: int,
    diagonalizer_algorithm: str,
    partitioning: bool,
    callbacks: Optional[Union[Callable[[GaussMNMF], None], List[Callable[[GaussMNMF], None]]]],
    normalization: Optional[Union[str, bool]],
    reset_kwargs: Dict[str, Any],
):
    if diagonalizer_algorithm in ["IP"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img[:n_channels], axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window=window, nperseg=n_fft, noverlap=n_fft - hop_length
    )

    if partitioning:
        with pytest.raises(AssertionError) as e:
            mnmf = FastGaussMNMF(
                n_basis,
                n_sources=n_sources,
                diagonalizer_algorithm=diagonalizer_algorithm,
                partitioning=partitioning,
                callbacks=callbacks,
                normalization=normalization,
                rng=rng,
            )

        assert str(e.value) == "partitioning function is not supported."
    else:
        mnmf = FastGaussMNMF(
            n_basis,
            n_sources=n_sources,
            diagonalizer_algorithm=diagonalizer_algorithm,
            partitioning=partitioning,
            callbacks=callbacks,
            normalization=normalization,
            rng=rng,
        )

        spectrogram_est = mnmf(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

        assert spectrogram_est.shape == (n_sources,) + spectrogram_mix.shape[1:]
        assert type(mnmf.loss[-1]) is float

        print(mnmf)



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.cacgmm import CACGMM

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.5
window = "hann"
n_fft = 512
hop_length = 256
n_bins = n_fft // 2 + 1
n_iter = 3
rng = np.random.default_rng(42)

parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_permutation_alignment = [
    "posterior_score",
    "amplitude_score",
    "amplitude_correlation",
]
parameters_cacgmm = [(2, 2, {}), (3, 2, {})]


@pytest.mark.parametrize("n_sources, n_channels, reset_kwargs", parameters_cacgmm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("permutation_alignment", parameters_permutation_alignment)
def test_CACGMM(
    n_sources: int,
    n_channels: int,
    callbacks: Optional[Union[Callable[[CACGMM], None], List[Callable[[CACGMM], None]]]],
    permutation_alignment: bool,
    reset_kwargs: Dict[str, Any],
):
    if n_sources < 4:
        sisec2010_tag = "dev1_female3"
    elif n_sources == 4:
        sisec2010_tag = "dev1_female4"
    else:
        raise ValueError("n_sources should be less than 5.")

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)
    waveform_mix = waveform_mix[:n_channels]

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window=window, nperseg=n_fft, noverlap=n_fft - hop_length
    )

    cacgmm = CACGMM(
        n_sources=n_sources,
        callbacks=callbacks,
        permutation_alignment=permutation_alignment,
        rng=rng,
    )

    spectrogram_est = cacgmm(spectrogram_mix, n_iter=n_iter, **reset_kwargs)

    assert spectrogram_est.shape == (n_sources,) + spectrogram_mix.shape[-2:]
    assert type(cacgmm.loss[-1]) is float

    # when posterior is not given
    _spectrogram_est = cacgmm.separate(spectrogram_mix)

    assert np.allclose(_spectrogram_est, spectrogram_est)

    print(cacgmm)



import functools
import warnings
from typing import Callable, Optional

import numpy as np

from ..special.flooring import max_flooring
from ..special.psd import to_psd as _to_psd

EPS = 1e-10


def to_psd(
    X: np.ndarray,
    axis1: int = -2,
    axis2: int = -1,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
) -> np.ndarray:
    r"""Ensure matrix to be positive semidefinite.

    Args:
        X (np.ndarray):
            A complex Hermitian matrix.
        axis1 (int):
            Axis to be used as first axis of 2D sub-arrays.
        axis2 (int):
            Axis to be used as second axis of 2D sub-arrays.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.

    Returns:
        Positive semidefinite matrix.
    """
    warnings.warn("Use ssspy.special.to_psd instead.", FutureWarning)

    return _to_psd(X, axis1=axis1, axis2=axis2, flooring_fn=flooring_fn)



import sys
from os.path import dirname, join, realpath

import numpy as np
import pytest

from ssspy.bss.fdica import AuxLaplaceFDICA, GradLaplaceFDICA, NaturalGradLaplaceFDICA

ssspy_tests_dir = dirname(dirname(dirname(realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.utils.dataset import load_regression_data  # noqa: E402

fdica_root = join(ssspy_tests_dir, "mock", "regression", "fdica")
n_sources = 2

parameters_is_holonomic = [True, False]
parameters_spatial_algorithm = ["IP1", "IP2"]


@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_laplace_fdica(is_holonomic: bool):
    if is_holonomic:
        root = join(fdica_root, "grad_laplace_fdica", "holonomic")
    else:
        root = join(fdica_root, "grad_laplace_fdica", "nonholonomic")

    npz_input, npz_target = load_regression_data(root=root)
    spectrogram_mix = npz_input["spectrogram"]
    spectrogram_trg = npz_target["spectrogram"]
    n_iter = npz_target["n_iter"].item()

    assert npz_target["is_holonomic"].item() == is_holonomic

    fdica = GradLaplaceFDICA(is_holonomic=is_holonomic)
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert np.allclose(spectrogram_est, spectrogram_trg)


@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_laplace_fdica(is_holonomic: bool):
    if is_holonomic:
        root = join(fdica_root, "natural_grad_laplace_fdica", "holonomic")
    else:
        root = join(fdica_root, "natural_grad_laplace_fdica", "nonholonomic")

    npz_input, npz_target = load_regression_data(root=root)
    spectrogram_mix = npz_input["spectrogram"]
    spectrogram_trg = npz_target["spectrogram"]
    n_iter = npz_target["n_iter"].item()

    assert npz_target["is_holonomic"].item() == is_holonomic

    fdica = NaturalGradLaplaceFDICA(is_holonomic=is_holonomic)
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert np.allclose(spectrogram_est, spectrogram_trg)


@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
def test_aux_laplace_fdica(spatial_algorithm: str):
    root = join(fdica_root, "aux_laplace_fdica", spatial_algorithm)
    npz_input, npz_target = load_regression_data(root=root)
    spectrogram_mix = npz_input["spectrogram"]
    spectrogram_trg = npz_target["spectrogram"]
    n_iter = npz_target["n_iter"].item()

    assert npz_target["spatial_algorithm"].item() == spatial_algorithm

    fdica = AuxLaplaceFDICA(spatial_algorithm=spatial_algorithm)
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert np.allclose(spectrogram_est, spectrogram_trg)



import pytest

from ssspy.utils.dataset import download_sample_speech_data

parameters_dataset = [
    (2, "dev1_female3"),
    (3, "dev1_female3"),
    (4, "dev1_female4"),
]
parameters_max_duration = [1.2]
parameters_conv = [True, False]


@pytest.mark.parametrize("n_sources, sisec2010_tag", parameters_dataset)
@pytest.mark.parametrize("max_duration", parameters_max_duration)
@pytest.mark.parametrize("conv", parameters_conv)
def test_conv_dataset(n_sources: int, sisec2010_tag: str, max_duration: int, conv: bool):
    waveform_src_img, sample_rate = download_sample_speech_data(
        sisec2010_root="./tests/.data/SiSEC2010",
        mird_root="./tests/.data/MIRD",
        n_sources=n_sources,
        sisec2010_tag=sisec2010_tag,
        max_duration=max_duration,
        conv=conv,
    )

    n_channels = n_sources

    assert waveform_src_img.shape == (n_channels, n_sources, int(sample_rate * max_duration))



import time

import numpy as np

from ssspy.linalg import cholesky2


def main():
    rng = np.random.default_rng(42)

    shape = (2, 2, 32)

    start = time.perf_counter()

    for _ in range(1000):
        x = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
        X = np.mean(x[:, :, np.newaxis, :] * x[:, np.newaxis, :, :].conj(), axis=-1)

        if False:
            _ = np.linalg.cholesky(X)
        else:
            _ = cholesky2(X)

    end = time.perf_counter()

    print(end - start)


if __name__ == "__main__":
    main()



from typing import Optional

import numpy as np
import pytest

from ssspy.algorithm import projection_back

parameters = [(2, 0), (3, 2), (2, None)]


@pytest.mark.parametrize("n_sources, reference_id", parameters)
def test_projection_back_demix_filter(n_sources: int, reference_id: Optional[int]):
    np.random.seed(111)

    n_channels = n_sources
    n_bins, n_frames = 17, 10

    spectrogram_mix = np.random.randn(n_channels, n_bins, n_frames) + 1j * np.random.randn(
        n_channels, n_bins, n_frames
    )
    demix_filter = np.random.randn(n_bins, n_sources, n_channels) + 1j * np.random.randn(
        n_bins, n_sources, n_channels
    )

    demix_filter_scaled = projection_back(demix_filter, reference_id=reference_id)

    spectrogram_est = demix_filter_scaled @ spectrogram_mix.transpose(1, 0, 2)

    if reference_id is None:
        spectrogram_est = spectrogram_est.transpose(0, 2, 1, 3)

        for _spectrogram_est in spectrogram_est:
            assert spectrogram_mix.shape == _spectrogram_est.shape
    else:
        spectrogram_est = spectrogram_est.transpose(1, 0, 2)

        assert spectrogram_mix.shape == spectrogram_est.shape


@pytest.mark.parametrize("n_sources, reference_id", parameters)
def test_projection_back_output(n_sources: int, reference_id: Optional[int]):
    np.random.seed(111)

    n_channels = n_sources
    n_bins, n_frames = 17, 10

    spectrogram_mix = np.random.randn(n_channels, n_bins, n_frames) + 1j * np.random.randn(
        n_channels, n_bins, n_frames
    )
    demix_filter = np.random.randn(n_bins, n_sources, n_channels) + 1j * np.random.randn(
        n_bins, n_sources, n_channels
    )
    spectrogram_est = demix_filter @ spectrogram_mix.transpose(1, 0, 2)
    spectrogram_est = spectrogram_est.transpose(1, 0, 2)

    spectrogram_est_scaled = projection_back(
        spectrogram_est, reference=spectrogram_mix, reference_id=reference_id
    )

    if reference_id is None:
        for _spectrogram_est_scaled in spectrogram_est_scaled:
            assert spectrogram_mix.shape == _spectrogram_est_scaled.shape
    else:
        assert spectrogram_mix.shape == spectrogram_est.shape



import numpy as np
import pytest

from ssspy.algorithm.permutation_alignment import (
    correlation_based_permutation_solver,
    score_based_permutation_solver,
)

rng = np.random.default_rng(0)

parameters_give_demixing_filter = [True, False]


@pytest.mark.parametrize("give_demixing_filter", parameters_give_demixing_filter)
def test_correlation_based_permutation_solver(give_demixing_filter: bool):
    n_sources = 3
    n_channels = n_sources
    n_bins, n_frames = 4, 16

    shape = (n_channels, n_bins, n_frames)
    mixture = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    shape = (n_bins, n_sources, n_channels)
    demix_filter = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    separated = demix_filter @ mixture.transpose(1, 0, 2)

    if give_demixing_filter:
        separated, demix_filter = correlation_based_permutation_solver(separated, demix_filter)

        assert demix_filter.shape == (n_bins, n_sources, n_channels)
    else:
        separated = correlation_based_permutation_solver(separated)

    assert separated.shape == (n_bins, n_sources, n_frames)


@pytest.mark.parametrize("give_demixing_filter", parameters_give_demixing_filter)
def test_score_based_permutation_solver(give_demixing_filter: bool):
    n_sources = 3
    n_channels = n_sources
    n_bins, n_frames = 4, 16

    shape = (n_channels, n_bins, n_frames)
    mixture = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    shape = (n_bins, n_sources, n_channels)
    demix_filter = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    separated = demix_filter @ mixture.transpose(1, 0, 2)

    if give_demixing_filter:
        separated, demix_filter = score_based_permutation_solver(separated, demix_filter)

        assert demix_filter.shape == (n_bins, n_sources, n_channels)
    else:
        separated = correlation_based_permutation_solver(separated)

    assert separated.shape == (n_bins, n_sources, n_frames)



import numpy as np

from ssspy.linalg import cbrt


def test_cbrt():
    rng = np.random.default_rng(0)
    n_bins = 8
    n_channels = 4

    # real number
    x = rng.standard_normal((n_bins, n_channels))
    y = cbrt(x)

    assert np.allclose(y**3, x)

    # complex number
    x = rng.standard_normal((n_bins, n_channels)) + 1j * rng.standard_normal((n_bins, n_channels))
    y = cbrt(x)

    assert np.allclose(y**3, x)



import pytest

from ssspy.bss._select_pair import combination_pair_selector, sequential_pair_selector

parameters_n_sources = [4]
parameters_step = [1, 2]
parameters_ascend = [True, False]


@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("step", parameters_step)
@pytest.mark.parametrize("ascend", parameters_ascend)
def test_sequential_pair_selector(n_sources: int, step: int, ascend: bool):
    with pytest.warns(UserWarning) as record:
        for m, n in sequential_pair_selector(n_sources, step=step, sort=ascend):
            if ascend:
                assert m < n

    assert len(record) == 1
    assert str(record[0].message) == "Use ssspy.utils.select_pair.sequential_pair_selector instead."


@pytest.mark.parametrize("n_sources", parameters_n_sources)
@pytest.mark.parametrize("ascend", parameters_ascend)
def test_combination_pair_selector(n_sources: int, ascend: bool):
    with pytest.warns(UserWarning) as record:
        for m, n in combination_pair_selector(n_sources, sort=ascend):
            if ascend:
                assert m < n

    assert len(record) == 1
    assert (
        str(record[0].message) == "Use ssspy.utils.select_pair.combination_pair_selector instead."
    )



from typing import Optional

import numpy as np
import pytest
import scipy.special

from ssspy.special import softmax

parameters_axis = [0, 1, (0, 2), None]


@pytest.mark.parametrize("axis", parameters_axis)
def test_logsumexp(axis: Optional[int]):
    rng = np.random.default_rng(0)

    n_sources, n_channels = 4, 3
    n_frames = 8
    shape = (n_sources, n_frames, n_channels, n_channels)

    X = rng.random(shape)

    Y = softmax(X, axis=axis)
    Y_scipy = scipy.special.softmax(X, axis=axis)

    assert np.allclose(Y, Y_scipy)



import numpy as np
import pytest
from scipy.linalg import sqrtm

from ssspy.linalg import gmeanmh

parameters_type = [1, 2, 3]


def gmeanmh_scipy(A: np.ndarray, B: np.ndarray, inverse="left") -> np.ndarray:
    def _sqrtm(X) -> np.ndarray:
        return np.stack([sqrtm(x) for x in X], axis=0)

    if inverse == "left":
        AB = np.linalg.solve(A, B)
        G = A @ _sqrtm(AB)
    elif inverse == "right":
        AB = np.linalg.solve(B, A)
        AB = AB.swapaxes(-2, -1).conj()
        G = _sqrtm(AB) @ B
    else:
        raise ValueError(f"Invalid inverse={inverse} is given.")

    return G


@pytest.mark.parametrize("type", parameters_type)
def test_gmean(type: int):
    rng = np.random.default_rng(0)
    size = (16, 32, 4, 1)

    def create_psd():
        x = rng.random(size) + 1j * rng.random(size)
        XX = x * x.transpose(0, 1, 3, 2).conj()

        return np.mean(XX, axis=0)

    A = create_psd()
    B = create_psd()

    G1 = gmeanmh(A, B, type=type)

    if type == 1:
        assert np.allclose(G1 @ np.linalg.inv(A) @ G1, B)
    elif type == 2:
        assert np.allclose(G1 @ A @ G1, B)
    elif type == 3:
        assert np.allclose(G1 @ np.linalg.inv(A) @ G1, np.linalg.inv(B))
    else:
        raise ValueError("Invalid type={} is given.".format(type))

    if type == 2:
        A = np.linalg.inv(A)
    elif type == 3:
        B = np.linalg.inv(B)

    G2 = gmeanmh_scipy(A, B, inverse="left")
    G3 = gmeanmh_scipy(A, B, inverse="right")

    assert np.allclose(G1, G2)
    assert np.allclose(G1, G3)



from typing import Tuple

import numpy as np
import pytest

from ssspy.bss._psd import to_psd
from ssspy.special import add_flooring

rng = np.random.default_rng(42)

parameters_shape = [(5, 2, 2), (3, 3)]
parameters_kwargs = [{}, {"flooring_fn": None}, {"flooring_fn": add_flooring}]


@pytest.mark.parametrize("shape", parameters_shape)
@pytest.mark.parametrize("kwargs", parameters_kwargs)
def test_to_psd_real(shape: Tuple[int], kwargs):
    X = rng.standard_normal(shape)
    X = X @ X.swapaxes(-1, -2)
    X = to_psd(X, **kwargs)
    eigvals = np.linalg.eigvalsh(X)

    assert np.all(X == X.swapaxes(-1, -2))
    assert np.min(eigvals) > 0


@pytest.mark.parametrize("shape", parameters_shape)
@pytest.mark.parametrize("kwargs", parameters_kwargs)
def test_to_psd_complex(shape: Tuple[int], kwargs):
    X = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    X = X @ X.swapaxes(-1, -2).conj()
    X = to_psd(X, **kwargs)
    eigvals = np.linalg.eigvalsh(X)

    assert np.all(X == X.swapaxes(-1, -2).conj())
    assert np.min(eigvals) > 0



import os
import sys
from typing import Callable, List, Optional, Union

import pytest

from ssspy.bss.base import IterativeMethodBase

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402

n_iter = 3

parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_record_loss = [True, False]


@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("record_loss", parameters_record_loss)
def test_iterative_method_base(
    callbacks: Optional[
        Union[Callable[[IterativeMethodBase], None], List[Callable[[IterativeMethodBase], None]]]
    ],
    record_loss: bool,
):
    method = IterativeMethodBase(callbacks=callbacks, record_loss=record_loss)

    with pytest.raises(NotImplementedError) as exc_info:
        method(n_iter=n_iter)

    assert exc_info.type is NotImplementedError



# Demixing filter updates
def update_by_ipa(
    demix_filter: np.ndarray,
    weighted_covariance: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    overwrite: bool = True,
) -> np.ndarray:
    r"""Update estimated spectrogram by iterative projection with adjustment (IPA).

    Args:
        demix_filter (numpy.ndarray):
            Demixing filters to be updated.
            The shape is (n_bins, n_sources, n_channels).
        weighted_covariance (numpy.ndarray):
            Weighted covariance matrix.
            The shape is (n_bins, n_sources, n_channels, n_channels).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        overwrite (bool):
            Overwrite ``demix_filter`` if ``overwrite=True``.
            Default: ``True``.

    Returns:
        numpy.ndarray of updated demixing filters of shape (n_sources, n_bins, n_frames).

    """
    if flooring_fn is None:
        flooring_fn = identity

    if overwrite:
        W = demix_filter
    else:
        W = demix_filter.copy()

    U = weighted_covariance

    n_bins, n_sources, n_channels = W.shape

    E = np.eye(n_sources, n_channels)

    for source_idx in range(n_sources):
        W_Hermite = W.transpose(0, 2, 1).conj()
        E_n_left, e_n, E_n_right = np.split(E, [source_idx, source_idx + 1], axis=-1)
        E_n = np.concatenate([E_n_left, E_n_right], axis=-1)
        U_tilde = W[:, np.newaxis, :, :] @ U @ W_Hermite[:, np.newaxis, :, :]
        U_tilde_n = U_tilde[:, source_idx, :, :]
        U_tilde_n = to_psd(U_tilde_n, axis1=-2, axis2=-1, flooring_fn=flooring_fn)
        U_tilde_n_inverse = np.linalg.inv(U_tilde_n)
        a_n = U_tilde[:, :, source_idx, source_idx]
        a_n = np.real(a_n)
        a_n = a_n @ E_n
        b_n = np.diagonal(U_tilde[:, :, source_idx, :], axis1=-2, axis2=-1)
        b_n = b_n @ E_n
        d_n = E_n.transpose(1, 0) @ U_tilde_n_inverse.conj()
        C_n = d_n @ E_n
        d_n = d_n[:, :, source_idx]

        Cd_n = np.linalg.solve(C_n, d_n)
        dCd_n = np.sum(d_n.conj() * Cd_n, axis=-1)
        dCd_n = np.real(dCd_n)
        eUe_n = U_tilde_n_inverse[:, source_idx, source_idx]
        eUe_n = np.real(eUe_n)
        z_n = eUe_n - dCd_n

        a_sqrt_n = np.sqrt(a_n)
        aa_n = a_sqrt_n[:, :, np.newaxis] * a_sqrt_n[:, np.newaxis, :]
        H_n = C_n / aa_n
        v_n = -b_n / a_sqrt_n - a_sqrt_n * Cd_n

        y_n = lqpqm2(
            H_n,
            v_n,
            z_n,
            flooring_fn=flooring_fn,
            singular_fn=lambda x: x < flooring_fn(0),
        )

        q_n = y_n / a_sqrt_n - b_n / a_n

        Eq_n = q_n.conj() @ E_n.transpose(1, 0)
        q_tilde_n = e_n.transpose(1, 0) - Eq_n

        Uq_n = np.linalg.solve(U_tilde_n, q_tilde_n)
        qUq_n = np.sum(q_tilde_n.conj() * Uq_n, axis=-1, keepdims=True)

        qUq_n = np.real(qUq_n)
        qUq_n = np.maximum(qUq_n, 0)
        denom = np.sqrt(qUq_n)
        denom = flooring_fn(denom)
        u_n = Uq_n / denom

        T_n = np.eye(n_sources, dtype=np.complex128)
        T_n = np.tile(T_n, reps=(n_bins, 1, 1))
        T_n[:, :, source_idx] = Eq_n
        T_n[:, source_idx, :] = u_n.conj()
        W[:] = T_n @ W

    return W



import numpy as np
import pytest

from ssspy.linalg import eigh

parameters_sources = [2, 5]
parameters_channels = [4, 3]
parameters_frames = [32, 16]
parameters_is_complex = [True, False]


@pytest.mark.parametrize("n_sources", parameters_sources)
@pytest.mark.parametrize("n_channels", parameters_channels)
@pytest.mark.parametrize("n_frames", parameters_frames)
@pytest.mark.parametrize("is_complex", parameters_is_complex)
def test_generalized_eigh(n_sources: int, n_channels: int, n_frames: int, is_complex: bool):
    np.random.seed(111)

    shape = (n_sources, n_channels, n_frames)

    if is_complex:
        a = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        b = np.random.randn(*shape) + 1j * np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :].conj(), axis=-1)
        B = np.mean(b[:, :, np.newaxis, :] * b[:, np.newaxis, :, :].conj(), axis=-1)
    else:
        a = np.random.randn(*shape)
        b = np.random.randn(*shape)
        A = np.mean(a[:, :, np.newaxis, :] * a[:, np.newaxis, :, :], axis=-1)
        B = np.mean(b[:, :, np.newaxis, :] * b[:, np.newaxis, :, :], axis=-1)

    lamb, Z = eigh(A, B)

    Lamb = lamb[..., np.newaxis] * np.eye(n_channels)
    BA = np.linalg.inv(B) @ A
    ZLZ = Z @ Lamb @ np.linalg.inv(Z)

    assert np.allclose(BA, ZLZ)



import numpy as np
import pytest

from ssspy.transform import whiten

parameters_batch_size = [1, 4]
parameters_n_channels = [2, 3]
parameters_whiten_real = [10, 20]
parameters_whiten_complex = [(2049, 8), (513, 12)]


@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_samples", parameters_whiten_real)
def test_whiten_real_2d(n_channels: int, n_samples: int):
    np.random.seed(111)

    input = np.random.randn(n_channels, n_samples)
    output = whiten(input)

    assert input.shape == output.shape

    covariance = output[:, np.newaxis, :] * output[np.newaxis, :, :]
    covariance = np.mean(covariance, axis=-1)
    eye = np.eye(n_channels)

    assert np.allclose(covariance, eye)


@pytest.mark.parametrize("batch_size", parameters_batch_size)
@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_samples", parameters_whiten_real)
def test_whiten_real_3d(batch_size: int, n_channels: int, n_samples: int):
    np.random.seed(111)

    input = np.random.randn(batch_size, n_channels, n_samples)
    output = whiten(input)

    assert input.shape == output.shape

    covariance = output[:, :, np.newaxis, :] * output[:, np.newaxis, :, :]
    covariance = np.mean(covariance, axis=-1)
    eye = np.eye(n_channels)

    assert np.allclose(covariance, eye)


@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_bins, n_frames", parameters_whiten_complex)
def test_whiten_complex_3d(n_channels: int, n_bins: int, n_frames: int):
    np.random.seed(111)

    real = np.random.randn(n_channels, n_bins, n_frames)
    imag = np.random.randn(n_channels, n_bins, n_frames)
    input = real + 1j * imag
    output = whiten(input)

    assert input.shape == output.shape

    covariance = output[:, np.newaxis, :, :] * output[np.newaxis, :, :, :].conj()
    covariance = np.mean(covariance, axis=-1)
    covariance = covariance.transpose(2, 0, 1)
    eye = np.eye(n_channels)
    eye = np.tile(eye, reps=(n_bins, 1, 1))

    assert np.allclose(covariance, eye)


@pytest.mark.parametrize("batch_size", parameters_batch_size)
@pytest.mark.parametrize("n_channels", parameters_n_channels)
@pytest.mark.parametrize("n_bins, n_frames", parameters_whiten_complex)
def test_whiten_complex_4d(batch_size: int, n_channels: int, n_bins: int, n_frames: int):
    np.random.seed(111)

    real = np.random.randn(batch_size, n_channels, n_bins, n_frames)
    imag = np.random.randn(batch_size, n_channels, n_bins, n_frames)
    input = real + 1j * imag
    output = whiten(input)

    assert input.shape == output.shape

    covariance = output[:, :, np.newaxis, :, :] * output[:, np.newaxis, :, :, :].conj()
    covariance = np.mean(covariance, axis=-1)
    covariance = covariance.transpose(0, 3, 1, 2)
    eye = np.eye(n_channels)
    eye = np.tile(eye, reps=(batch_size, n_bins, 1, 1))

    assert np.allclose(covariance, eye)



import numpy as np
import pytest

from ssspy.bss._solve_permutation import correlation_based_permutation_solver

rng = np.random.default_rng(0)

parameters_give_demixing_filter = [True, False]


@pytest.mark.parametrize("give_demixing_filter", parameters_give_demixing_filter)
def test_correlation_based_permutation_solver(give_demixing_filter: bool):
    n_sources = 3
    n_channels = n_sources
    n_bins, n_frames = 4, 16

    shape = (n_channels, n_bins, n_frames)
    mixture = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    shape = (n_bins, n_sources, n_channels)
    demix_filter = rng.standard_normal(shape) + 1j * rng.standard_normal(shape)
    separated = demix_filter @ mixture.transpose(1, 0, 2)

    with pytest.warns(UserWarning) as record:
        if give_demixing_filter:
            separated, demix_filter = correlation_based_permutation_solver(separated, demix_filter)
            assert demix_filter.shape == (n_bins, n_sources, n_channels)
        else:
            separated = correlation_based_permutation_solver(separated)

            assert separated.shape == (n_bins, n_sources, n_frames)

    assert len(record) == 1
    assert (
        str(record[0].message)
        == "Use ssspy.algorithm.permutation_alignment.correlation_based_permutation_solver instead."
    )



import numpy as np


def decorrelate(input: np.ndarray) -> np.ndarray:
    U, _, V_Hermite = np.linalg.svd(input)
    output = U @ V_Hermite

    return output


def decorrelate_by_gram_schmidt(input: np.ndarray) -> np.ndarray:
    if src_idx > 0:
        W_n = W[:src_idx]  # (src_idx - 1, n_channels)
        scale = np.sum(W_n * w_n, axis=-1, keepdims=True)
        w_n = w_n - np.sum(scale * W_n, axis=0)



import numpy as np


def inv2(X: np.ndarray) -> np.ndarray:
    r"""Compute the (multiplicative) inverse of a 2x2 matrix.

    Args:
        X (numpy.ndarray):
            2x2 matrix to be inverted. The shape is (\*, 2, 2).

    Returns:
        numpy.ndarray:
            (Multiplicative) inverse of the matrix X.

    Examples:
        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.linalg import inv2
            >>> X = np.array([[0, 1], [2, 3]])
            >>> X_inv = inv2(X)
            >>> np.allclose(X @ X_inv, np.eye(2))
            True
            >>> np.allclose(X_inv @ X, np.eye(2))
            True

        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.linalg import inv2
            >>> X = np.array([[[0, 1], [2, 3]], [[4, 5], [6, 7]]])
            >>> inv2(X)
            array([[[-1.5,  0.5],
                    [ 1. , -0. ]],

                [[-3.5,  2.5],
                    [ 3. , -2. ]]])
    """
    shape = X.shape

    assert shape[-2:] == (2, 2), "2x2 matrix is expected, but given shape of {}.".format(shape)

    a = X[..., 0, 0]
    b = X[..., 0, 1]
    c = X[..., 1, 0]
    d = X[..., 1, 1]

    det = a * d - b * c

    X_adj = np.stack([d, -b, -c, a], axis=-1)
    X_adj = X_adj.reshape(shape[:-2] + (2, 2))
    X_inv = X_adj / det[..., np.newaxis, np.newaxis]

    return X_inv



import numpy as np
import pytest

from ssspy.linalg import inv2

parameters_sources = [2, 5]


@pytest.mark.parametrize("n_sources", parameters_sources)
def test_inv2(n_sources: int):
    np.random.seed(111)

    shape = (n_sources, 2, 2)

    A = np.random.randn(*shape) + 1j * np.random.randn(*shape)
    B = inv2(A)

    assert np.allclose(A @ B, np.eye(2))



import numpy as np

from .eigh import eigh


def gmeanmh(A: np.ndarray, B: np.ndarray, type: int = 1) -> np.ndarray:
    r"""Compute the geometric mean of complex Hermitian \
    (conjugate symmetric) or real symmetric matrices.

    The geometric mean of :math:`\boldsymbol{A}` and :math:`\boldsymbol{B}`
    is defined as follows [#bhatia2009positive]_:

    .. math::
        \boldsymbol{A}\#\boldsymbol{B}
        &= \boldsymbol{A}^{1/2}
        (\boldsymbol{A}^{-1/2}\boldsymbol{B}\boldsymbol{A}^{-1/2})^{1/2}
        \boldsymbol{A}^{1/2} \\
        &= \boldsymbol{A}(\boldsymbol{A}^{-1}\boldsymbol{B})^{1/2} \\
        &= (\boldsymbol{A}\boldsymbol{B}^{-1})^{1/2}\boldsymbol{B}.

    This is a solution of the following equation for
    complex Hermitian or real symmetric matrices,
    :math:`\boldsymbol{A}`, :math:`\boldsymbol{B}`, and :math:`\boldsymbol{X}`:

    .. math::
        \boldsymbol{X}\boldsymbol{A}^{-1}\boldsymbol{X} = \boldsymbol{B}.

    .. note::
        In this toolkit, :math:`\boldsymbol{A}\#\boldsymbol{B}` is computed by
        :math:`\boldsymbol{B}(\boldsymbol{B}^{-1}\boldsymbol{A})^{1/2}`
        in terms of computational speed.
        Note that :math:`\boldsymbol{A}\#\boldsymbol{B}` is equal to
        :math:`\boldsymbol{B}\#\boldsymbol{A}`.
        For comparison of computational time, see https://github.com/tky823/ssspy/issues/210.

    .. note::
        :math:`(\boldsymbol{B}^{-1}\boldsymbol{A})^{1/2}` is computed by
        generalized eigendecomposition.
        Let :math:`\lambda` and :math:`z` be the eigenvalue and eigenvector of
        the generalized eigenproblem :math:`\boldsymbol{Az}=\lambda\boldsymbol{Bz}`.
        Then, :math:`(\boldsymbol{B}^{-1}\boldsymbol{A})^{1/2}` is computed by
        :math:`\boldsymbol{Z}\boldsymbol{\Lambda}^{1/2}\boldsymbol{Z}^{-1}`,
        where the main diagonals of :math:`\boldsymbol{\Lambda}` are :math:`\lambda` s
        and the columns of :math:`\boldsymbol{Z}` are :math:`\boldsymbol{z}` s.

    Args:
        A (numpy.ndarray):
            A complex Hermitian matrix with shape of (\*, n_channels, n_channels).
        B (numpy.ndarray):
            A complex Hermitian matrix with shape of (\*, n_channels, n_channels).
        type (int):
            This value specifies the type of geometric mean.
            Only ``1``, ``2``, and ``3`` are supported.

            - When ``type=1``, return :math:`\boldsymbol{A}\#\boldsymbol{B}`.
            - When ``type=2``, return :math:`\boldsymbol{A}^{-1}\#\boldsymbol{B}`.
            - When ``type=3``, return :math:`\boldsymbol{A}\#\boldsymbol{B}^{-1}`.

    Returns:
        Geometric mean of matrices with shape of (\*, n_channels, n_channels).

    .. [#bhatia2009positive] R. Bhatia,
        "Positive definite matrices,"
        Princeton university press, 2009.
    """  # noqa: W605
    lamb, Z = eigh(A, B, type=type)
    lamb = np.sqrt(lamb)
    Lamb = lamb[..., np.newaxis] * np.eye(Z.shape[-1])
    ZLZ = Z @ Lamb @ np.linalg.inv(Z)

    if type == 1:
        BA = ZLZ
        G = B @ BA
    elif type == 2:
        AB = ZLZ
        G = np.linalg.inv(A) @ AB
    elif type == 3:
        BA = ZLZ
        G = np.linalg.inv(B) @ BA
    else:
        raise ValueError("Invalid type={} is given.".format(type))

    return G



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest

from ssspy.bss.ica import (
    FastICA,
    GradICA,
    GradICABase,
    GradLaplaceICA,
    NaturalGradICA,
    NaturalGradLaplaceICA,
)

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.5
n_iter = 3

parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_is_holonomic = [True, False]
parameters_grad_ica = [
    (2, {}),
    (3, {"demix_filter": -np.eye(3)}),
]
parameters_fast_ica = [
    (2, {}),
    (3, {"demix_filter": -np.eye(3)}),
]


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_ica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_ica_base(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradICA], None], List[Callable[[GradICA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    def contrast_fn(x):
        return np.log(1 + np.exp(x))

    def score_fn(x):
        return 1 / (1 + np.exp(-x))

    ica = GradICABase(contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks)

    print(ica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_ica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_ica(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradICA], None], List[Callable[[GradICA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=False,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    def contrast_fn(x):
        return np.log(1 + np.exp(x))

    def score_fn(x):
        return 1 / (1 + np.exp(-x))

    ica = GradICA(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )
    waveform_est = ica(waveform_mix, n_iter=n_iter)

    assert waveform_mix.shape == waveform_est.shape
    assert type(ica.loss[-1]) is float

    print(ica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_ica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_ica(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradICA], None], List[Callable[[GradICA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=False,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    def contrast_fn(x):
        return np.log(1 + np.exp(x))

    def score_fn(x):
        return 1 / (1 + np.exp(-x))

    ica = NaturalGradICA(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )
    waveform_est = ica(waveform_mix, n_iter=n_iter)

    assert waveform_mix.shape == waveform_est.shape
    assert type(ica.loss[-1]) is float

    print(ica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_ica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_laplace_ica(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradICA], None], List[Callable[[GradICA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=False,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    ica = GradLaplaceICA(callbacks=callbacks, is_holonomic=is_holonomic)
    waveform_est = ica(waveform_mix, n_iter=n_iter)

    assert waveform_mix.shape == waveform_est.shape
    assert type(ica.loss[-1]) is float

    print(ica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_ica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_laplace_ica(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradICA], None], List[Callable[[GradICA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=False,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    ica = NaturalGradLaplaceICA(callbacks=callbacks, is_holonomic=is_holonomic)
    waveform_est = ica(waveform_mix, n_iter=n_iter)

    assert waveform_mix.shape == waveform_est.shape
    assert type(ica.loss[-1]) is float

    print(ica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_fast_ica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_fast_ica(
    n_sources: int,
    callbacks: Optional[Union[Callable[[FastICA], None], List[Callable[[FastICA], None]]]],
    reset_kwargs: Dict[Any, Any],
):
    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=False,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    def contrast_fn(x):
        return np.log(1 + np.exp(x))

    def score_fn(x):
        return 1 / (1 + np.exp(-x))

    def d_score_fn(x):
        s = 1 / (1 + np.exp(-x))
        return s * (1 - s)

    ica = FastICA(
        contrast_fn=contrast_fn, score_fn=score_fn, d_score_fn=d_score_fn, callbacks=callbacks
    )
    waveform_est = ica(waveform_mix, n_iter=n_iter)

    assert waveform_mix.shape == waveform_est.shape
    assert type(ica.loss[-1]) is float

    print(ica)



import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pytest
import scipy.signal as ss

from ssspy.bss.fdica import (
    AuxFDICA,
    AuxLaplaceFDICA,
    GradFDICA,
    GradFDICABase,
    GradLaplaceFDICA,
    NaturalGradFDICA,
    NaturalGradLaplaceFDICA,
)

ssspy_tests_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(ssspy_tests_dir)

from dummy.callback import DummyCallback, dummy_function  # noqa: E402
from dummy.utils.dataset import download_sample_speech_data  # noqa: E402

max_duration = 0.5
n_fft = 512
hop_length = 256
n_bins = n_fft // 2 + 1
n_iter = 3

parameters_callbacks = [None, dummy_function, [DummyCallback(), dummy_function]]
parameters_is_holonomic = [True, False]
parameters_scale_restoration = [True, False, "projection_back", "minimal_distortion_principle"]
parameters_spatial_algorithm = ["IP", "IP1", "IP2"]
parameters_grad_fdica = [
    (2, {}),
    (
        3,
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
]
parameters_aux_fdica = [
    (2, {}),
    (
        3,
        {"demix_filter": np.tile(-np.eye(3, dtype=np.complex128), reps=(n_bins, 1, 1))},
    ),
]


@pytest.mark.parametrize("callbacks", parameters_callbacks)
def test_grad_fdica_base(
    callbacks: Optional[Union[Callable[[GradFDICA], None], List[Callable[[GradFDICA], None]]]],
):
    np.random.seed(111)

    def contrast_fn(y):
        return 2 * np.abs(y)

    def score_fn(y):
        denominator = np.maximum(np.abs(y), 1e-10)
        return y / denominator

    fdica = GradFDICABase(contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks)

    print(fdica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_fdica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_fdica(
    n_sources: int,
    callbacks: Optional[Union[Callable[[GradFDICA], None], List[Callable[[GradFDICA], None]]]],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y):
        return 2 * np.abs(y)

    def score_fn(y):
        denominator = np.maximum(np.abs(y), 1e-10)
        return y / denominator

    fdica = GradFDICA(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(fdica.loss[-1]) is float

    print(fdica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_fdica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_fdica(
    n_sources: int,
    callbacks: Optional[
        Union[Callable[[NaturalGradFDICA], None], List[Callable[[NaturalGradFDICA], None]]]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y):
        return 2 * np.abs(y)

    def score_fn(y):
        denominator = np.maximum(np.abs(y), 1e-10)
        return y / denominator

    fdica = NaturalGradFDICA(
        contrast_fn=contrast_fn, score_fn=score_fn, callbacks=callbacks, is_holonomic=is_holonomic
    )
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(fdica.loss[-1]) is float

    print(fdica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_aux_fdica)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_aux_fdica(
    n_sources: int,
    spatial_algorithm: str,
    callbacks: Optional[Union[Callable[[AuxFDICA], None], List[Callable[[AuxFDICA], None]]]],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[Any, Any],
):
    if spatial_algorithm in ["IP"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    def contrast_fn(y):
        return 2 * np.abs(y)

    def d_contrast_fn(y):
        return 2 * np.ones_like(y)

    fdica = AuxFDICA(
        spatial_algorithm=spatial_algorithm,
        contrast_fn=contrast_fn,
        d_contrast_fn=d_contrast_fn,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(fdica.loss[-1]) is float

    print(fdica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_fdica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_grad_laplace_fdica(
    n_sources: int,
    callbacks: Optional[
        Union[Callable[[GradLaplaceFDICA], None], List[Callable[[GradLaplaceFDICA], None]]]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    fdica = GradLaplaceFDICA(callbacks=callbacks, is_holonomic=is_holonomic)
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(fdica.loss[-1]) is float

    print(fdica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_grad_fdica)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("is_holonomic", parameters_is_holonomic)
def test_natural_grad_laplace_fdica(
    n_sources: int,
    callbacks: Optional[
        Union[
            Callable[[NaturalGradLaplaceFDICA], None],
            List[Callable[[NaturalGradLaplaceFDICA], None]],
        ]
    ],
    is_holonomic: bool,
    reset_kwargs: Dict[Any, Any],
):
    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    fdica = NaturalGradLaplaceFDICA(callbacks=callbacks, is_holonomic=is_holonomic)
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(fdica.loss[-1]) is float

    print(fdica)


@pytest.mark.parametrize("n_sources, reset_kwargs", parameters_aux_fdica)
@pytest.mark.parametrize("spatial_algorithm", parameters_spatial_algorithm)
@pytest.mark.parametrize("callbacks", parameters_callbacks)
@pytest.mark.parametrize("scale_restoration", parameters_scale_restoration)
def test_aux_laplace_fdica(
    n_sources: int,
    spatial_algorithm: str,
    callbacks: Optional[
        Union[Callable[[AuxLaplaceFDICA], None], List[Callable[[AuxLaplaceFDICA], None]]]
    ],
    scale_restoration: Union[str, bool],
    reset_kwargs: Dict[Any, Any],
):
    if spatial_algorithm in ["IP"] and not pytest.run_redundant:
        pytest.skip(reason="Need --run-redundant option to run.")

    np.random.seed(111)

    waveform_src_img, _ = download_sample_speech_data(
        n_sources=n_sources,
        sisec2010_tag="dev1_female3",
        max_duration=max_duration,
        conv=True,
    )
    waveform_mix = np.sum(waveform_src_img, axis=1)  # (n_channels, n_samples)

    _, _, spectrogram_mix = ss.stft(
        waveform_mix, window="hann", nperseg=n_fft, noverlap=n_fft - hop_length
    )

    fdica = AuxLaplaceFDICA(
        spatial_algorithm=spatial_algorithm,
        callbacks=callbacks,
        scale_restoration=scale_restoration,
    )
    spectrogram_est = fdica(spectrogram_mix, n_iter=n_iter)

    assert spectrogram_mix.shape == spectrogram_est.shape
    assert type(fdica.loss[-1]) is float

    print(fdica)



from . import permutation_alignment
from .minimal_distortion_principle import minimal_distortion_principle
from .projection_back import projection_back

__all__ = ["permutation_alignment", "minimal_distortion_principle", "projection_back"]

PROJECTION_BACK_KEYWORDS = ["projection_back", "projection-back", "PB"]
MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS = [
    "minimal_distortion_principle",
    "minimal-distortion-principle",
    "MDP",
]



import os
import struct

import numpy as np


def save_invalid_wavfile(
    path: str,
    invalid_riff: bool = False,
    invalid_ftype: bool = False,
    invalid_fmt_chunk_marker: bool = False,
    invalid_fmt_chunk_size: bool = False,
    invalid_fmt: bool = False,
    invalid_byte_rate: bool = False,
    invalid_data_chunk_marker: bool = False,
) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)

    n_channels = 1
    sample_rate = 16000
    bits_per_sample = 16
    duration = 5
    byte_rate = (bits_per_sample * sample_rate * n_channels) // 8
    block_align = byte_rate // sample_rate
    total_file_size = byte_rate * duration + 44

    rng = np.random.default_rng(42)
    num_frames = sample_rate * duration
    bytes_per_sample = block_align // n_channels
    vmax = 2 ** (bits_per_sample - 1)

    valid_file_size = 0

    with open(path, mode="wb") as f:
        if invalid_riff:
            data = b"RIFX"
        else:
            data = b"RIFF"

        f.write(data)
        valid_file_size += 4

        data = struct.pack("<I", total_file_size - 4 - 4)
        f.write(data)
        valid_file_size += 4

        if invalid_ftype:
            data = b"wave"
        else:
            data = b"WAVE"

        f.write(data)
        valid_file_size += 4

        if invalid_fmt_chunk_marker:
            data = b"FMT "
        else:
            data = b"fmt "

        f.write(data)
        valid_file_size += 4

        if invalid_fmt_chunk_size:
            data = struct.pack("<I", 15)
        else:
            data = struct.pack("<I", 16)

        f.write(data)
        valid_file_size += 4

        if invalid_fmt:
            data = struct.pack("<H", 0)
        else:
            data = struct.pack("<H", 1)

        f.write(data)
        valid_file_size += 2

        if invalid_byte_rate:
            data = struct.pack(
                "<HIIHH", n_channels, sample_rate, byte_rate + 1, block_align, bits_per_sample
            )
        else:
            data = struct.pack(
                "<HIIHH", n_channels, sample_rate, byte_rate, block_align, bits_per_sample
            )

        f.write(data)
        valid_file_size += 2 + 4 + 4 + 2 + 2

        if invalid_data_chunk_marker:
            data = b"DATA"
        else:
            data = b"data"

        f.write(data)
        valid_file_size += 4

        data_chunk_size = num_frames * n_channels * block_align
        data = struct.pack("<I", data_chunk_size)
        f.write(data)
        valid_file_size += 4

        waveform = rng.integers(-vmax, vmax, size=(num_frames,), dtype=f"<i{bytes_per_sample}")
        data = waveform.view("b").data
        f.write(data)
        valid_file_size += byte_rate * duration

        assert valid_file_size == total_file_size



import numpy as np


def softmax(X: np.ndarray, axis: int = None) -> np.ndarray:
    r"""Compute softmax values.

    Args:
        X (np.ndarray):
            Elements to compute softmax.
        axis (int or tuple[int], optional):
            Axis or axes over which the sum is performed.
            Default: ``None``.

    Returns:
        np.ndarray of softmax values.

    Examples:

        .. code-block:: python

            >>> import numpy as np

            >>> X = np.array([[1, 2, 3], [4, 5, 6]])
            >>> softmax(X, axis=0)
            array([[0.04742587, 0.04742587, 0.04742587],
                [0.95257413, 0.95257413, 0.95257413]])
            >>> softmax(X, axis=1)
            array([[0.09003057, 0.24472847, 0.66524096],
                [0.09003057, 0.24472847, 0.66524096]])
    """
    vmax = np.max(X, axis=axis, keepdims=True)
    Y = X - vmax
    exp = np.exp(Y)
    v = exp / np.sum(exp, axis=axis, keepdims=True)

    return v
